import tempfile
import unittest
from pathlib import Path

from tira.tira_run import guess_dataset


class TestGuessDataset(unittest.TestCase):
    def test_empty_dir_does_not_yield_dataset(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            actual = guess_dataset(tmp_dir)
            self.assertIsNone(actual)

    def test_ambiguous_dataset_does_not_yield_guess(self):
        guess_from_dir = Path(__file__).parent.parent / "resources" / "valid-ir-metadata-multiple-datasets-ambiguous"
        actual = guess_dataset(guess_from_dir)
        self.assertIsNone(actual)

    def test_non_ambiguous_dataset_does_yield_guess(self):
        expected = "learned-sparse-retrieval-20250926-training"
        guess_from_dir = Path(__file__).parent.parent / "resources" / "valid-ir-metadata-multiple-datasets"
        actual = guess_dataset(guess_from_dir)
        self.assertEqual(actual, expected)

    def test_non_ambiguous_dataset_does_yield_guess_hidden(self):
        expected = "other-dataset"
        guess_from_dir = Path(__file__).parent.parent / "resources" / "valid-ir-metadata-multiple-datasets-hidden"
        actual = guess_dataset(guess_from_dir)
        self.assertEqual(actual, expected)



import os
import tempfile
import unittest
from pathlib import Path
from zipfile import ZipFile

from tira.rest_api_client import Client


class CodeSubmissionTest(unittest.TestCase):
    def test_code_submission_fails_if_code_not_in_version_control(self):
        tira = Client()
        with tempfile.TemporaryDirectory() as tmp_file:
            with self.assertRaises(ValueError):
                tira.submit_code(Path(tmp_file), "wows-eval", dry_run=True)

    def test_code_submission_fails_for_dirty_git_repo(self):
        tira = Client()
        with tempfile.TemporaryDirectory() as tmp_file:
            with ZipFile(Path("tests") / "resources" / "example-git-repositories.zip", "r") as zip_ref:
                zip_ref.extractall(tmp_file)
            with self.assertRaises(ValueError):
                tira.submit_code(Path(tmp_file) / "git-repo-dirty" / "some-directory", "wows-eval", dry_run=True)

    def test_code_submission_works(self):
        tira = Client(tira_cache_dir="./tests/resources/local_cached_zip")
        expected_code_files = ["some-directory/.gitignore", "some-directory/Dockerfile", "some-directory/script.sh"]

        with tempfile.TemporaryDirectory() as tmp_file:
            with ZipFile(Path("tests") / "resources" / "example-git-repositories.zip", "r") as zip_ref:
                zip_ref.extractall(tmp_file)

            os.chmod(str(Path(tmp_file) / "git-repo-clean" / "some-directory" / "script.sh"), 0o0766)
            actual = tira.submit_code(
                Path(tmp_file) / "git-repo-clean" / "some-directory", "task-does-not-exist", dry_run=True
            )

        zipObj = ZipFile(actual["code"])
        files_in_zip = [i.filename for i in zipObj.infolist()]

        self.assertEqual({"origin": "foo"}, actual["remotes"])
        self.assertEqual("976c6949b9992aabc785ccb8544652dc3b149fb5", actual["commit"])
        self.assertEqual("main", actual["active_branch"])
        self.assertTrue(actual["image"].startswith("some-directory"))

        self.assertEqual(files_in_zip, expected_code_files)



import unittest

from tira.tira_run import guess_system_details, guess_vm_id_of_user

from ..format_check import JSONL_OUTPUT_VALID


class TestGuessDetails(unittest.TestCase):
    def test_guess_vm_id_01(self):
        expected = "hello-world"
        actual = guess_vm_id_of_user(rest_client=None, tira_task_id=None, tira_vm_id="hello-world")
        self.assertEqual(expected, actual)

    def test_guess_vm_id_02(self):
        expected = "foo"
        actual = guess_vm_id_of_user(rest_client=None, tira_task_id=None, tira_vm_id="foo")
        self.assertEqual(expected, actual)

    def test_guess_system_details_01(self):
        expected = {"tag": "hello-world"}
        actual = guess_system_details(directory=None, system="hello-world")
        self.assertEqual(expected, actual)

    def test_guess_system_details_02(self):
        expected = {"tag": "foo"}
        actual = guess_system_details(directory=None, system="foo")
        self.assertEqual(expected, actual)

    def test_guess_system_details_03(self):
        actual = guess_system_details(
            directory=JSONL_OUTPUT_VALID.parent / "longeval-ir-metadata" / "ir-metadata.yml", system=None
        )
        self.assertEqual("ows_bm25_bo1_keyqueries", actual["tag"])
        self.assertIn("A keyquerey approach that reformulates", actual["description"])



import json
import tempfile
import unittest
from pathlib import Path

from approvaltests import verify_as_json

from tira.check_format import QueryProcessorFormat, check_format, lines_if_valid

from . import _ERROR, _OK, EMPTY_OUTPUT, IR_QUERY_OUTPUT


class TestQueryProcessorFormat(unittest.TestCase):
    def test_valid_query_processor_format(self):
        """Test that a valid query processor output directory passes validation."""
        expected = [_OK, "The jsonl file has the correct format."]
        actual = check_format(IR_QUERY_OUTPUT, "query-processor")
        self.assertEqual(expected, actual)

    def test_lines_for_valid_query_processor_format(self):
        """Test that a valid query processor output directory passes validation."""
        actual = lines_if_valid(IR_QUERY_OUTPUT, "query-processor")
        verify_as_json(actual)

    def test_error_message_on_empty_output(self):
        """Test error message when no jsonl file is found."""
        expected = [_ERROR, "No unique *.jsonl file was found, only the files ['.gitkeep'] were available."]
        actual = check_format(EMPTY_OUTPUT, "query-processor")
        self.assertEqual(expected, actual)

    def test_query_processor_missing_query_id(self):
        """Test validation fails when an id is missing."""
        validator = QueryProcessorFormat()
        with self.assertRaises(ValueError) as context:
            validator.fail_if_json_line_is_not_valid(
                {
                    "wrong_id": "123",
                    "originalQuery": "test query",
                    "segmentationApproach": "test-approach",
                    "segmentation": ["segmen", "tation"],
                }
            )

        self.assertIn("At least one of ", str(context.exception))

    def test_query_processor_unequal_queries(self):
        """Test validation fails both possible id names are present and unequal."""
        validator = QueryProcessorFormat()
        with self.assertRaises(ValueError) as context:
            validator.fail_if_json_line_is_not_valid(
                {
                    "query_id": "123",
                    "qid": "456",
                    "originalQuery": "test query",
                    "segmentationApproach": "test-approach",
                    "segmentation": ["stuff", "more stuff"],
                }
            )

        self.assertIn("they must be equal", str(context.exception))

    def test_query_processor_duplicate_query_ids(self):
        """Test validation fails when duplicate query IDs are found."""
        # Create a temporary directory with a JSONL file containing duplicate query IDs
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir_path = Path(temp_dir)

            # Create JSONL file with duplicate query IDs
            jsonl_file = temp_dir_path / "queries.jsonl"
            with open(jsonl_file, "w") as f:
                f.write(
                    json.dumps({"qid": "123", "originalQuery": "first query", "segmentation": ["part1", "part2"]})
                    + "\n"
                )
                f.write(
                    json.dumps(
                        {"qid": "456", "originalQuery": "second query", "segmentation": ["segment1", "segment2"]}
                    )
                    + "\n"
                )
                f.write(
                    json.dumps(
                        {
                            "qid": "123",  # Duplicate ID
                            "originalQuery": "duplicate query",
                            "segmentation": ["foo", "bar"],
                        }
                    )
                    + "\n"
                )

            # Verify that validation fails
            validator = QueryProcessorFormat()
            with self.assertRaises(ValueError) as context:
                list(validator.yield_next_entry(temp_dir_path))

            self.assertIn("Query ID 123 is not unique", str(context.exception))



import os
import tempfile
import unittest
import warnings
from pathlib import Path

import pandas as pd
import pyterrier as pt
import pyterrier_alpha as pta

from tira.pyterrier_util import TiraSourceTransformer

from . import load_artifact


class TestWorkingArtifacts(unittest.TestCase):
    def test_bm25_artifact(self):
        expected = {
            "qid": "51",
            "docno": "S885c6b4f-Ad3259e81",
            "rank": 1,
            "score": 26.07818108898493,
            "name": "pyterrier.default_pipelines.wmodel_batch_retrieve",
        }
        bm25 = load_artifact(
            "tira:argsme/2020-04-01/touche-2021-task-1/tira-ir-starter/BM25 (tira-ir-starter-pyterrier)"
        )

        run = bm25.transform(pd.DataFrame([{"qid": "51"}]))
        actual = run.iloc[0].to_dict()
        self.assertEqual(expected, actual)

    def test_hyb_a_query_segmentation_artifact(self):
        expected = ["should", "blood donations", "be financially compensated"]
        hyb_a = load_artifact("tira:argsme/2020-04-01/touche-2021-task-1/ows/query-segmentation-hyb-a")

        self.assertIsNotNone(hyb_a)
        actual = hyb_a.transform(pd.DataFrame([{"qid": "53"}])).iloc[0].to_dict()
        self.assertIn("segmentation", actual)
        self.assertEqual(expected, actual["segmentation"])

    def test_doc_t5_query_artifact(self):
        expected = "including fluid motion"
        docT5Query = load_artifact("tira:cranfield/seanmacavaney/DocT5Query (Local)")

        self.assertIsNotNone(docT5Query)
        actual = docT5Query.transform(pd.DataFrame([{"docno": "873"}])).iloc[0].to_dict()
        self.assertIn("text", actual)
        self.assertIn(expected, actual["text"])

    def test_index_artifact(self):
        import pyterrier as pt

        index = load_artifact("tira:cranfield/tira-ir-starter/Index (tira-ir-starter-pyterrier)")

        self.assertIn("IndexRef", str(index))
        run = pt.terrier.Retriever(index).search("chemical reactions")
        actual = run.iloc[0].to_dict()
        self.assertIsNotNone(actual)
        self.assertEqual("488", actual["docno"])

    def test_artifact_is_tira_source_transformer(self):
        """Test that retriever-artifacts loaded from TIRA URLs are TiraSourceTransformer instances"""
        # Test with a BM25 artifact
        art = pta.Artifact.from_url("tira:cranfield/tira-ir-starter/BM25 (tira-ir-starter-pyterrier)")
        self.assertIsInstance(art, TiraSourceTransformer)

    def test_non_artifact_is_regular_source_transformer(self):
        """Test that local run files create regular SourceTransformer, not TiraSourceTransformer"""
        # Create a temporary run file
        with tempfile.NamedTemporaryFile(mode="w", suffix=".txt", delete=False) as f:
            # Write sample TREC run format data
            f.write("1 Q0 doc1 1 10.0 system\n")
            f.write("1 Q0 doc2 2 9.0 system\n")
            f.write("2 Q0 doc3 1 8.0 system\n")
            temp_run_file = f.name

        try:
            # Load as regular PyTerrier dataframe and transformer
            live_df = pt.io.read_results(temp_run_file)
            live_tr = pt.Transformer.from_df(live_df)

            # Should be SourceTransformer but NOT TiraSourceTransformer
            self.assertIsInstance(live_tr, pt.transformer.SourceTransformer)
            self.assertNotIsInstance(live_tr, TiraSourceTransformer)
        finally:
            # Clean up temp file
            os.unlink(temp_run_file)

    def test_on_column_mismatch_warn(self):
        """Test that on_column_mismatch='warn' produces warnings for extra query columns"""
        art = pta.Artifact.from_url("tira:cranfield/tira-ir-starter/BM25 (tira-ir-starter-pyterrier)")

        # Create topics with extra query rewrite columns
        topics = pd.DataFrame(
            [{"qid": "1", "query": "test query", "query_0": "rewritten query 1", "query_1": "rewritten query 2"}]
        )

        # Should warn by default
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            art.transform(topics)

            # Check that a warning was issued
            self.assertTrue(len(w) > 0)
            warning_msg = str(w[0].message)
            self.assertIn("cannot process rewritten query columns", warning_msg)
            self.assertIn("will be ignored during retrieval", warning_msg)

    def test_on_column_mismatch_error(self):
        """Test that on_column_mismatch='error' raises an error for extra query columns"""
        # Create artifact with error mode
        art = pta.Artifact.from_url("tira:cranfield/tira-ir-starter/BM25 (tira-ir-starter-pyterrier)")
        art.on_column_mismatch = "error"

        # Create topics with extra query rewrite columns
        topics = pd.DataFrame([{"qid": "1", "query": "test query", "query_0": "rewritten query 1"}])

        # Should raise ValueError
        with self.assertRaises(ValueError) as cm:
            art.transform(topics)

        error_msg = str(cm.exception)
        self.assertIn("may not be the intended behavior", error_msg)

    def test_on_column_mismatch_ignore(self):
        """Test that on_column_mismatch='ignore' silently ignores extra query columns"""
        # Create artifact with ignore mode
        art = pta.Artifact.from_url("tira:cranfield/tira-ir-starter/BM25 (tira-ir-starter-pyterrier)")
        art.on_column_mismatch = "ignore"

        # Create topics with extra query rewrite columns
        topics = pd.DataFrame([{"qid": "1", "query": "test query", "query_0": "rewritten query 1"}])

        # Should not warn or error
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            art.transform(topics)

            # Check that no warnings were issued about column mismatch
            column_warnings = [warning for warning in w if "ignores rewritten queries" in str(warning.message)]
            self.assertEqual(len(column_warnings), 0)

    def test_on_column_mismatch_default_behavior(self):
        """Test that the default on_column_mismatch behavior is 'warn'"""
        art = pta.Artifact.from_url("tira:cranfield/tira-ir-starter/BM25 (tira-ir-starter-pyterrier)")

        # Check that default is 'warn'
        self.assertEqual(art.on_column_mismatch, "warn")

    def test_env_var_changes_default(self):
        """
        If TIRA_ARTIFACT_ON_REWRITE is set, every artifact loaded via pta.Artifact.from_url
        should inherit that mode automatically.
        """
        import os
        from unittest.mock import patch

        # env-var override using patch
        with patch.dict(os.environ, {"TIRA_ARTIFACT_ON_COLUMN_MISMATCH": "error"}):
            # Load artifact - should inherit the environment variable setting
            art = pta.Artifact.from_url("tira:cranfield/tira-ir-starter/BM25 (tira-ir-starter-pyterrier)")

            self.assertEqual(art.on_column_mismatch, "error")

    def test_get_metadata_transformer_artifact(self):
        """Test that transformer artifacts have get_metadata() method that returns correct data"""
        art = pta.Artifact.from_url("tira:cranfield/tira-ir-starter/BM25 (tira-ir-starter-pyterrier)")

        # Should have get_metadata method
        self.assertTrue(hasattr(art, "get_metadata"))
        self.assertTrue(callable(art.get_metadata))

        # Should return metadata dict
        metadata = art.get_metadata()
        self.assertIsInstance(metadata, dict)
        self.assertEqual(metadata.get("type"), "tira")
        self.assertEqual(metadata.get("format"), "pt_transformer")

    def test_get_metadata_index_artifact(self):
        """Test that index artifacts have get_metadata() method that returns correct data"""
        art = pta.Artifact.from_url("tira:cranfield/tira-ir-starter/Index (tira-ir-starter-pyterrier)")

        # Should have get_metadata method
        self.assertTrue(hasattr(art, "get_metadata"))
        self.assertTrue(callable(art.get_metadata))

        # Should return metadata dict
        metadata = art.get_metadata()
        self.assertIsInstance(metadata, dict)
        self.assertEqual(metadata.get("type"), "tira")
        self.assertEqual(metadata.get("format"), "pt_index_transformer")

    def test_get_metadata_query_transformer_artifact(self):
        """Test that query transformer artifacts have get_metadata() method that returns correct data"""
        art = pta.Artifact.from_url("tira:cranfield/qspell/hunspell")

        # Should have get_metadata method
        self.assertTrue(hasattr(art, "get_metadata"))
        self.assertTrue(callable(art.get_metadata))

        # Should return metadata dict
        metadata = art.get_metadata()
        self.assertIsInstance(metadata, dict)
        self.assertEqual(metadata.get("type"), "tira")
        self.assertEqual(metadata.get("format"), "pt_query_transformer")

    def test_get_metadata_document_transformer_artifact(self):
        """Test that document transformer artifacts have get_metadata() method that returns correct data"""
        art = pta.Artifact.from_url("tira:cranfield/seanmacavaney/DocT5Query (Local)")

        # Should have get_metadata method
        self.assertTrue(hasattr(art, "get_metadata"))
        self.assertTrue(callable(art.get_metadata))

        # Should return metadata dict
        metadata = art.get_metadata()
        self.assertIsInstance(metadata, dict)
        self.assertEqual(metadata.get("type"), "tira")
        self.assertEqual(metadata.get("format"), "pt_document_transformer")

    def test_local_bm25_artifact_from_file(self):
        expected = {
            "qid": "3",
            "docno": "doc-3",
            "rank": 1,
            "score": 1.0,
            "name": "tag",
        }
        bm25 = load_artifact(
            "tira:" + str((Path(__file__).parent.parent / "resources" / "ranking-outputs" / "run.txt").absolute())
        )

        run = bm25.transform(pd.DataFrame([{"qid": "3"}]))
        print(run)
        actual = run.iloc[0].to_dict()
        self.assertEqual(expected, actual)

    def test_local_bm25_artifact_from_dir(self):
        expected = {
            "qid": "3",
            "docno": "doc-3",
            "rank": 1,
            "score": 1.0,
            "name": "tag",
        }
        bm25 = load_artifact("tira:" + str((Path(__file__).parent.parent / "resources" / "ranking-outputs").absolute()))

        run = bm25.transform(pd.DataFrame([{"qid": "3"}]))
        print(run)
        actual = run.iloc[0].to_dict()
        self.assertEqual(expected, actual)



import gzip
import json
import os
import shutil
import uuid
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

from tira.check_format import report_valid_formats
from tira.third_party_integrations import temporary_directory
from tira.tirex import IRDS_TO_TIREX_DATASET, TIREX_ARTIFACT_DEBUG_URL

if TYPE_CHECKING:
    import numpy as np
    import pandas as pd

    from tira.pyterrier_util import TiraApplyFeatureTransformer, TiraFullRankTransformer

    from .rest_api_client import Client

OUTPUT_FORMAT_TO_PT_FORMAT: Dict[str, str] = {
    "run.txt": "pt_transformer",
    "query-processor": "pt_query_transformer",
    "terrier-index": "pt_index_transformer",
    "document-processor": "pt_document_transformer",
}


class PyTerrierIntegration:
    def __init__(self, tira_client: "Client") -> None:
        self.tira_client = tira_client
        self.pd = tira_client.pd
        self.irds_docker_image = "webis/tira-application:0.0.36"

    def retriever(self, approach: str, dataset: "Optional[str]" = None) -> "TiraFullRankTransformer":
        from tira.pyterrier_util import TiraFullRankTransformer

        input_dir = self.ensure_dataset_is_cached(dataset, dataset)
        return TiraFullRankTransformer(approach, self.tira_client, input_dir)

    def ensure_dataset_is_cached(self, irds_dataset_id: "Optional[str]", dataset: "Optional[str]" = None) -> Path:
        from tira.io_utils import run_cmd

        cache_dir = Path(self.tira_client.tira_cache_dir) / "pyterrier" / str(irds_dataset_id)
        full_rank_data = cache_dir / "full-rank"
        truth_data = cache_dir / "truth-data"
        irds_cache = cache_dir / "irds-cache"

        if os.path.isfile(full_rank_data / "documents.jsonl"):
            return full_rank_data

        Path(full_rank_data).mkdir(parents=True, exist_ok=True)
        Path(truth_data).mkdir(parents=True, exist_ok=True)
        Path(irds_cache).mkdir(parents=True, exist_ok=True)

        run_cmd(
            [
                "docker",
                "run",
                "-v",
                f"{irds_cache}:/root/.ir_datasets/:rw",
                "-v",
                f"{full_rank_data}:/output/:rw",
                "-v",
                "{truth_data}:/truth/:rw",
                "--entrypoint",
                "/irds_cli.sh",
                self.irds_docker_image,
                "--output_dataset_path",
                "/output",
                "--ir_datasets_id",
                irds_dataset_id,
                "--output_dataset_truth_path",
                "/truth",
            ]
        )

        return Path(full_rank_data)

    def create_rerank_file(
        self, run_df: "pd.DataFrame" = None, run_file: "Optional[Path]" = None, irds_dataset_id: "Optional[str]" = None
    ) -> Path:
        from tira.io_utils import run_cmd
        from tira.third_party_integrations import persist_and_normalize_run

        if run_df is None and run_file is None:
            raise ValueError("Please pass either run_df or run_file")

        if run_file is not None:
            return run_file

        run_file = temporary_directory()
        Path(run_file).mkdir(parents=True, exist_ok=True)

        if "text" not in run_df.columns and "body" not in run_df.columns:
            if not irds_dataset_id:
                raise ValueError(f"Please pass a irds_dataset_id. Got {irds_dataset_id}.")
            persist_and_normalize_run(run_df, "system-is-ignored", run_file)

            cache_dir = self.tira_client.tira_cache_dir + "/pyterrier/" + irds_dataset_id
            irds_cache = cache_dir + "/irds-cache/"

            run_cmd(
                [
                    "docker",
                    "run",
                    "-v",
                    irds_cache + ":/root/.ir_datasets/:rw",
                    "-v",
                    f"{run_file}:/output/:rw",
                    "--entrypoint",
                    "/irds_cli.sh",
                    self.irds_docker_image,
                    "--output_dataset_path",
                    "/output",
                    "--ir_datasets_id",
                    irds_dataset_id,
                    "--rerank",
                    "/output",
                ]
            )
        else:
            with gzip.open(run_file / "rerank.jsonl.gz", "wt") as f:
                for _, i in run_df.iterrows():
                    i = i.to_dict()

                    for k in ["original_query", "original_document"]:
                        if k not in i:
                            i[k] = {}

                    if "text" not in i and "body" in i:
                        i["text"] = i["body"]

                    if "text" not in i:
                        raise ValueError(f'I expect a field "text", but only found fields {i.keys()}.')

                    f.write(json.dumps(i) + "\n")

        return Path(run_file)

    def index(self, approach, dataset):
        """
        Load an PyTerrier index from TIRA.
        """
        import pyterrier as pt

        from tira.ir_datasets_util import translate_irds_id_to_tirex

        ret = self.tira_client.get_run_output(approach, translate_irds_id_to_tirex(dataset)) / "index"
        return pt.IndexFactory.of(os.path.abspath(str(ret)))

    def from_submission(self, approach, dataset=None, datasets=None, file_to_re_rank=None):
        if self.__is_re_ranker(approach):
            return self.from_retriever_submission(approach, dataset, datasets=datasets, file_to_re_rank=file_to_re_rank)
        else:
            from tira.pyterrier_util import TiraRerankingTransformer

            return TiraRerankingTransformer(approach, self.tira_client, dataset, datasets)

    def __is_re_ranker(self, approach):
        if self.tira_client.input_run_in_sandbox(approach):
            # If the input run is in the sandbox, everything behaves as a re-ranker
            return True

        _, team, software = approach.split("/")
        software = self.tira_client.public_system_details(team, software)
        return software.get("ir_re_ranker", False)

    def from_retriever_submission(self, approach, dataset, previous_stage=None, datasets=None, file_to_re_rank=None):
        from tira.ir_datasets_loader import IrDatasetsLoader
        from tira.pyterrier_util import TiraSourceTransformer

        if file_to_re_rank:
            re_rank_dir = Path(self.tira_client.tira_cache_dir) / "extracted_datasets" / str(uuid.uuid4())
            log_file = Path(f"{self.tira_client.tira_cache_dir}/.archived/local-executions.jsonl")
            re_rank_dir = IrDatasetsLoader().reformat_to_re_rank_dataset(
                Path(file_to_re_rank), dataset, re_rank_dir, log_file
            )
            dataset = str(Path(re_rank_dir))

        ret = self.pd.from_retriever_submission(approach, dataset, previous_stage, datasets)
        return TiraSourceTransformer(ret)

    def transform_queries(self, approach, dataset, format: str = "query-processor", prefix: str = ""):
        from pyterrier.apply import generic

        ret = self.pd.transform_queries(approach, dataset, format)
        cols = [i for i in ret.columns if i not in ["qid"]]
        ret = {str(i["qid"]): i for _, i in ret.iterrows()}

        def __transform_df(df):
            df = df.copy()
            for col in cols:
                df[prefix + col] = df["qid"].apply(lambda i: ret[str(i)][col])
            return df

        return generic(__transform_df)

    def transform_documents(self, approach, dataset, format: str = "document-processor", prefix: str = ""):
        from pyterrier.apply import generic

        ret = self.pd.transform_documents(approach, dataset, format)
        cols = [i for i in ret.columns if i not in ["docno"]]
        ret = {str(i["docno"]): i for _, i in ret.iterrows()}

        def __transform_df(df):
            for col in cols:
                df[prefix + col] = df["docno"].apply(lambda i: ret[str(i)][col])
            return df

        return generic(__transform_df)

    @staticmethod
    def _get_features_from_row(row: Any, cols: List[str], map_features: Any = None) -> "np.ndarray":
        import numpy as np

        res: List = []

        for c in cols:
            if map_features is not None and c in map_features:
                f = map_features[c](row[c])
            else:
                f = row[c]

            if isinstance(f, (list, np.ndarray)):
                res.extend(f)
            else:
                res.append(f)

        return np.array(res)

    def _features_transformer(
        self, run: "pd.DataFrame", id_col: str, name: str, feature_selection: Any = None, map_features: Any = None
    ) -> "TiraApplyFeatureTransformer":
        from tira.pyterrier_util import TiraApplyFeatureTransformer

        cols = [col for col in run.columns if col != id_col]
        if feature_selection is not None:
            cols = [col for col in cols if col in feature_selection]

        mapping = {str(row[id_col]): self._get_features_from_row(row, cols, map_features) for _, row in run.iterrows()}

        return TiraApplyFeatureTransformer(mapping, (id_col,), name)

    def doc_features(
        self,
        approach: str,
        dataset: str,
        format: str = "document-processor",
        feature_selection: Any = None,
        map_features: Any = None,
    ) -> "TiraApplyFeatureTransformer":
        run = self.pd.transform_documents(approach, dataset, format)

        return self._features_transformer(run, "docno", "doc_features", feature_selection, map_features)

    def query_features(
        self,
        approach,
        dataset: "Optional[str]" = None,
        format: str = "query-processor",
        feature_selection: Any = None,
        map_features: Any = None,
    ) -> "TiraApplyFeatureTransformer":
        run = self.pd.transform_queries(approach, dataset, format)

        return self._features_transformer(run, "qid", "query_features", feature_selection, map_features)

    def reranker(self, approach, irds_id=None):
        from tira.pyterrier_util import TiraLocalExecutionRerankingTransformer

        return TiraLocalExecutionRerankingTransformer(approach, self.tira_client, irds_id=irds_id)


class PyTerrierAnceIntegration:
    """
    The pyterrier_ance integration to re-use cached ANCE indices. Wraps https://github.com/terrierteam/pyterrier_ance
    """

    def __init__(self, tira_client):
        self.tira_client = tira_client

    def ance_retrieval(self, dataset: str):
        """Load a cached pyterrier_ance.ANCEIndexer submitted as workshop-on-open-web-search/ows/pyterrier-anceindex
        from tira.

        References (for citation):
            https://arxiv.org/pdf/2007.00808.pdf
            https://github.com/microsoft/ANCE/

        Args:
            dataset (str): the dataset id, either an tira or ir_datasets id.

        Returns:
            pyterrier_ance.ANCERetrieval: the ANCE index.
        """
        from tira.ir_datasets_util import translate_irds_id_to_tirex

        ance_index = (
            Path(
                self.tira_client.get_run_output(
                    "ir-lab-sose-2024/ows/pyterrier-anceindex", translate_irds_id_to_tirex(dataset)
                )
            )
            / "anceindex"
        )
        ance_checkpoint = self.tira_client.load_resource("Passage_ANCE_FirstP_Checkpoint.zip")
        import pyterrier_ance

        return pyterrier_ance.ANCERetrieval(ance_checkpoint, ance_index)


class PyTerrierSpladeIntegration:
    """The pyt_splade integration to re-use cached Splade indices. Wraps https://github.com/cmacdonald/pyt_splade"""

    def __init__(self, tira_client):
        self.tira_client = tira_client

    def splade_index(self, dataset: str, approach: str = "workshop-on-open-web-search/naverlabseurope/Splade (Index)"):
        """
        Load a cached pyt_splade index submitted as the passed approach (default
        'workshop-on-open-web-search/naverlabseurope/Splade (Index)') from tira.

        References (for citation):
            https://github.com/naver/splade?tab=readme-ov-file#cite-scroll
            ToDo: Ask Thibault what to cite.

        Args:
            dataset (str): the dataset id, either an tira or ir_datasets id.
            approach (str, optional): the approach id, defaults
                'workshop-on-open-web-search/naverlabseurope/Splade (Index)'.

        Returns:
            The PyTerrier index suitable for retrieval.
        """
        import pyterrier as pt

        from tira.ir_datasets_util import translate_irds_id_to_tirex

        ret = (
            Path(
                self.tira_client.get_run_output(
                    "ir-lab-sose-2024/naverlabseurope/Splade (Index)", translate_irds_id_to_tirex(dataset)
                )
            )
            / "spladeindex"
        )
        return pt.IndexFactory.of(os.path.abspath(ret))


def pt_document_transformer(path):
    import pyterrier as pt

    if not pt.started():
        pt.init()
    from .rest_api_client import Client

    original_path = path
    path = Path(path)
    if (path / "output").exists():
        path = path / "output"

    transformer = PyTerrierIntegration(Client()).transform_documents(path, None)
    transformer.artifact_path = original_path
    return _add_metadata_support(transformer, original_path)


def pt_query_transformer(path):
    import pyterrier as pt

    if not pt.started():
        pt.init()
    from .rest_api_client import Client

    original_path = path
    path = Path(path)
    if (path / "output").exists():
        path = path / "output"

    transformer = PyTerrierIntegration(Client()).transform_queries(path, None)
    transformer.artifact_path = original_path
    return _add_metadata_support(transformer, original_path)


def pt_index_transformer(path):
    import pyterrier as pt

    if not pt.started():
        pt.init()

    index_ref = pt.IndexRef.of(str((Path(path) / "output" / "index").resolve().absolute()))
    index_ref.artifact_path = path
    return _add_metadata_support(index_ref, path)


def pt_transformer(path):
    import pyterrier as pt

    from .pyterrier_util import TiraSourceTransformer

    if not pt.started():
        pt.init()
    # TODO hacked for the moment, in reality, we must delegate to the classes above.

    original_path = path
    run_path = os.path.join(path, "output", "run.txt")
    df = pt.io.read_results(run_path)

    mode = os.getenv("TIRA_ARTIFACT_ON_COLUMN_MISMATCH", "warn").lower()
    if mode not in ["warn", "error", "ignore"]:
        raise ValueError(
            f"Invalid TIRA_ARTIFACT_ON_COLUMN_MISMATCH value: {mode}. Expected 'warn', 'error', or 'ignore'."
        )

    transformer = TiraSourceTransformer(df, on_column_mismatch=mode)
    transformer.artifact_path = original_path
    return _add_metadata_support(transformer, original_path)


def _extract_dataset_team_approach(url: str) -> Tuple[str, str, str, str]:
    dataset_id = None

    if len(url) < 5:
        raise ValueError(f"Invalid tira url. I expected 'tira:<IR-DATASETS-ID>/<TEAM>/<APPROACH>'. But got '{url}'.")

    for irds_id, tira_dataset_id in IRDS_TO_TIREX_DATASET.items():
        if url.startswith(irds_id):
            dataset_id = tira_dataset_id
            url = url.replace(irds_id, "ir-benchmarks")
            break

    if dataset_id is None:
        raise ValueError(
            f"Invalid tira url. I expected 'tira:<IR-DATASETS-ID>/<TEAM>/<APPROACH>'. But could not find a ir-dataset."
            f"I got '{url}'. Please see {TIREX_ARTIFACT_DEBUG_URL} for an overview of all available dataset ids."
        )

    if len(url.split("/")) != 3:
        raise ValueError(
            f"Invalid tira url. I expected 'tira:<IR-DATASETS-ID>/<TEAM>/<APPROACH>'."
            f"I found the dataset {irds_id} but have no team and/or approach."
        )
    team, approach = url.split("/")[1:]

    return dataset_id, url, team, approach


def _download_artifact_from_tira(url: str) -> Tuple[Path, str]:
    from tira.rest_api_client import Client

    dataset_id, url, team, approach = _extract_dataset_team_approach(url)

    tira = Client()

    all_systems = tira.archived_json_response("/v1/systems/all")
    pt_format = None

    for system in all_systems:
        if team == system["team"] and approach == system["name"]:
            if "verified_outputs" not in system:
                raise ValueError("Fooo")

            if dataset_id not in system["verified_outputs"]:
                raise ValueError("Fooo")

            for k in system["verified_outputs"][dataset_id]:
                pt_format = OUTPUT_FORMAT_TO_PT_FORMAT[k]

    if pt_format is None:
        raise ValueError(
            f"No submission '{approach}' by team '{team}' is publicly available in TIRA."
            f"Please see all public submissions at " + TIREX_ARTIFACT_DEBUG_URL
        )

    ret = tira.get_run_output(url, dataset_id)
    return Path(ret).parent, pt_format


def _local_artifact_from_tira(file_or_directory: Path) -> Tuple[Path, str]:
    ret = temporary_directory() / "output"

    if file_or_directory.is_file():
        ret.mkdir()
        shutil.copy(file_or_directory, ret)
    else:
        shutil.copytree(file_or_directory, ret)
    valid_formats = report_valid_formats(ret)
    pt_formats = [v for k, v in OUTPUT_FORMAT_TO_PT_FORMAT.items() if k in valid_formats]

    if len(pt_formats) != 1:
        raise ValueError("I do not know in which format the artifact is.")

    return (ret.parent, pt_formats[0])


def pt_artifact_entrypoint(url) -> str:
    url = url.netloc + url.path

    if url and Path(url).exists():
        ret, pt_format = _local_artifact_from_tira(Path(url))
    else:
        ret, pt_format = _download_artifact_from_tira(url)

    if not (ret / "pt_meta.json").is_file():
        with open(ret / "pt_meta.json", "w") as f:
            f.write(json.dumps({"type": "tira", "format": pt_format}))

    return str(ret.absolute())


def _add_metadata_support(transformer, artifact_path):
    """Add get_metadata() method to any PyTerrier transformer or IndexRef"""

    def get_metadata():
        """
        Fetch and read the pt_meta.json file for this artifact.

        Returns:
            dict: The parsed JSON content of the pt_meta.json file, or None if not found.
        """
        if not hasattr(transformer, "artifact_path") or not transformer.artifact_path:
            print("No artifact_path set for this transformer. Cannot read metadata.")
            return None

        meta_path = Path(transformer.artifact_path) / "pt_meta.json"

        if not meta_path.exists():
            return None

        try:
            with open(meta_path, "r") as f:
                return json.load(f)
        except (json.JSONDecodeError, IOError) as e:
            import warnings

            warnings.warn(f"Could not read pt_meta.json: {e}", RuntimeWarning)
            return None

    # Monkey-patch the method onto the transformer
    transformer.get_metadata = get_metadata
    return transformer



import gzip
import json
import os
import re
import xml.dom.minidom
from collections.abc import Iterable
from enum import Enum
from glob import glob
from pathlib import Path
from typing import Any, Dict, Optional, Sequence, Union


class FormatMsgType(Enum):
    OK = 0
    WARN = 1
    ERROR = 2


_fmt = FormatMsgType


def fmt_message(message: str, level: _fmt) -> str:
    """
    Prints a formatted log message with a symbol indicating the status.

    Parameters:
    - message (str): The log message to display.
    - level (_fmt): The level of the message; can be _fmt.OK, _fmt.WARN, _fmt.ERROR.
    """
    symbols = {
        _fmt.OK: "\033[92m\u2713\033[0m",  # Green check mark
        _fmt.WARN: "\033[93m" + b"\xe2\x9a\xa0".decode("utf-8") + "\033[0m",  # Yellow warning
        _fmt.ERROR: "\033[91m" + b"\xe2\x9c\x96".decode("utf-8") + "\033[0m",  # Red cross
    }

    symbol = symbols[level]
    return f"{symbol} {message}"


def log_message(message: str, level: _fmt):
    """
    Prints a formatted log message with a symbol indicating the status.

    Parameters:
    - message (str): The log message to display.
    - level (_fmt): The level of the message; can be _fmt.OK, _fmt.WARN, _fmt.ERROR.
    """
    print(fmt_message(message, level))


CONF_REQUIRED_FIELDS = "required_fields"
CONF_MINIMUM_LINES = "minimum_lines"
CONF_ID_FIELD = "id_field"
CONF_VALUE_FIELD = "value_field"
CONF_MAX_SIZE_MB = "max_size_mb"
CONF_SPOT_CHECK = "spot_check"

CONFIGURATION_FIELDS = {
    CONF_REQUIRED_FIELDS: "foo",
    CONF_MINIMUM_LINES: "foo",
    CONF_ID_FIELD: "foo",
    CONF_VALUE_FIELD: "foo",
}


class FormatBase:
    def __init__(self):
        self.max_size_mb = 75
        self.spot_check = None

    def yield_next_entry(self, f: Path):
        try:
            f_size = f.stat().st_size
        except:
            raise ValueError(f"File {f} is not readable.")

        if f_size > self.max_size():
            raise ValueError(f"File {f} is too large (size {f_size} exceeds configured {self.max_size()}).")

        if str(f).endswith(".gz"):
            with gzip.open(f, "rt") as o:
                for line in o:
                    yield line.rstrip("\n")
        else:
            with open(f, "r") as o:
                for line in o:
                    yield line.rstrip("\n")

    def apply_configuration_and_throw_if_invalid(self, configuration: "Optional[dict[str, Any]]"):
        if configuration and hasattr(configuration, "__iter__") and CONF_MAX_SIZE_MB in configuration:
            self.max_size_mb = configuration[CONF_MAX_SIZE_MB]

        if configuration and hasattr(configuration, "__iter__") and CONF_SPOT_CHECK in configuration:
            self.spot_check = configuration[CONF_SPOT_CHECK]

    def all_lines(self, f: Path):
        return list(self.yield_next_entry(f))

    def max_size(self):
        return self.max_size_mb * 1024 * 1024

    def check_format(self, run_output: Path):
        return [_fmt.ERROR, "not implemented"]


class LearnedSparseRetrievalInputs(FormatBase):
    """Checks if a given output is a valid learned sparese retrieval input."""

    def apply_configuration_and_throw_if_invalid(self, configuration: "Optional[dict[str, Any]]"):
        self.docs = DocumentProcessorFormat()
        self.docs.apply_configuration_and_throw_if_invalid(configuration)

        self.queries = QueryProcessorFormat()
        self.queries.apply_configuration_and_throw_if_invalid(configuration)

    def check_format(self, run_output: Path):
        l, m = self.docs.check_format(run_output / "corpus.jsonl.gz")
        if l != _fmt.OK:
            return l, m

        l, m = self.queries.check_format(run_output / "queries.jsonl")
        if l != _fmt.OK:
            return l, m

        # TODO: IrMetadataFormat

        return [_fmt.OK, "The dataset is in the format for the lsr-benchmark."]


class RunFormat(FormatBase):
    """Checks if a given output is a valid run file."""

    def check_format(self, run_output: Path):
        if (run_output / "run.txt").exists() and (run_output / "run.txt.gz").exists():
            msg = f"Found multiple run.txt or run.txt.gz files: {os.listdir(run_output)} ."
            return [_fmt.ERROR, msg]
        if not (run_output / "run.txt").exists() and not (run_output / "run.txt.gz").exists():
            msg = "No file run.txt or run.txt.gz was found."
            try:
                msg += " Only the files " + str(os.listdir(run_output)) + " were available."
            except:
                pass
            return [_fmt.ERROR, msg]

        try:
            # maximum size of 10MB
            num_lines = 0
            lines = []
            for line in self.yield_next_entry(run_output):
                if self.spot_check and num_lines > self.spot_check:
                    break
                lines.append(line)
                num_lines += 1
        except Exception as e:
            return [_fmt.ERROR, e.args[0]]

        if len(lines) < 10:
            return [_fmt.ERROR, f"The run file contains only {len(lines)} lines, this is likely an error."]

        query_to_docs = {}
        for line in lines:
            if line["qid"] not in query_to_docs:
                query_to_docs[line["qid"]] = set()
            if line["docno"] in query_to_docs[line["qid"]]:
                return [
                    _fmt.ERROR,
                    f'The run file has duplicate documents: the document with id "{line["docno"]}" appears multiple times for query "{line["qid"]}".',
                ]
            query_to_docs[line["qid"]].add(line["docno"])

        if len(query_to_docs.keys()) < 3:
            return [_fmt.ERROR, f"The run file has only {len(query_to_docs)} queries which is likely an error."]

        return [_fmt.OK, "The run.txt file has the correct format."]

    def parse_single_line(self, line):
        cols = re.split("\\s+", line)
        if len(cols) != 6:
            raise ValueError(
                f'Invalid line in the run file, expected 6 columns, but found a line "{line}" with {len(cols)} columns.'
            )
        qid, q0, docno, rank, score, system = cols
        try:
            rank = int(rank)
        except ValueError:
            raise ValueError(f"Could not parse the rank. Got {rank}. Line: {line}")

        try:
            score = float(score)
        except ValueError:
            raise ValueError(f"Could not parse the rank. Got {rank}. Line: {line}")

        return {"qid": qid, "q0": q0, "docno": docno, "rank": rank, "score": score, "system": system}

    def yield_next_entry(self, run_output):
        if (run_output / "run.txt").exists():
            file_path = run_output / "run.txt"
        elif (run_output / "run.txt.gz").exists():
            file_path = run_output / "run.txt.gz"
        else:
            raise ValueError("Could not find a file run.txt or run.txt.gz")

        for line in super().yield_next_entry(file_path):
            yield self.parse_single_line(line.strip())


class QrelFormat(FormatBase):
    """Checks if a given output is a valid qrel file."""

    def check_format(self, run_output: Path):
        if (run_output / "qrels.txt").exists() and (run_output / "qrels.txt.gz").exists():
            msg = f"Found multiple qrels.txt or qrels.txt.gz files: {os.listdir(run_output)} ."
            return [_fmt.ERROR, msg]
        if not (run_output / "qrels.txt").exists() and not (run_output / "qrels.txt.gz").exists():
            msg = "No unique qrels.txt file was found, only the files "
            msg += str(os.listdir(run_output)) + " were available."
            return [_fmt.ERROR, msg]

        try:
            # maximum size of 10MB
            lines = self.all_lines(run_output)
        except Exception as e:
            return [_fmt.ERROR, e.args[0]]

        if len(lines) < 10:
            return [_fmt.ERROR, f"The run file contains only {len(lines)} lines, this is likely an error."]

        query_to_docs = {}
        for line in lines:
            if line["qid"] not in query_to_docs:
                query_to_docs[line["qid"]] = set()
            if line["docno"] in query_to_docs[line["qid"]]:
                return [
                    _fmt.ERROR,
                    f'The qrel file has duplicate documents: the document with id "{line["docno"]}" appears multiple times for query "{line["qid"]}".',
                ]
            query_to_docs[line["qid"]].add(line["docno"])

        if len(query_to_docs.keys()) < 3:
            return [_fmt.ERROR, f"The run file has only {len(query_to_docs)} queries which is likely an error."]

        return [_fmt.OK, "The qrels are valid."]

    def all_lines(self, run_output):
        if (run_output / "qrels.txt").exists():
            ret = [i.strip() for i in super().all_lines(run_output / "qrels.txt")]
        elif (run_output / "qrels.txt.gz").exists():
            ret = [i.strip() for i in super().all_lines(run_output / "qrels.txt.gz")]
        else:
            raise ValueError("Could not find a file qrels.txt or qrels.txt.gz")

        ret_parsed = []
        for l in ret:
            l = l.strip()
            cols = re.split("\\s+", l)
            if len(cols) != 4:
                return [
                    _fmt.ERROR,
                    f'Invalid line in the qrel file, expected 4 columns, but found a line "{l}" with {len(cols)} columns.',
                ]
            qid, q0, docno, rel = cols
            try:
                rel = int(rel)
            except:
                raise ValueError(f"I expected that the relevance is an integer, got {rel} in line {l}")
            ret_parsed.append({"qid": qid, "q0": q0, "docno": docno, "rel": rel})
        return ret_parsed


class KeyValueFormatBase(FormatBase):
    def apply_configuration_and_throw_if_invalid(self, configuration: "Optional[dict[str, Any]]"):
        super().apply_configuration_and_throw_if_invalid(configuration)

        id_field = None
        value_field = None
        required_fields = set()
        minimum_lines = 1

        if configuration and hasattr(configuration, "__iter__"):
            if CONF_REQUIRED_FIELDS in configuration:
                required_fields = set(configuration[CONF_REQUIRED_FIELDS])
            if CONF_MINIMUM_LINES in configuration:
                minimum_lines = int(configuration[CONF_MINIMUM_LINES])

            if CONF_ID_FIELD in configuration:
                id_field = configuration[CONF_ID_FIELD]

            if CONF_VALUE_FIELD in configuration:
                value_field = configuration[CONF_VALUE_FIELD]

        self.required_fields = required_fields
        self.minimum_lines = minimum_lines
        self.id_field = id_field
        self.value_field = value_field

        if (self.id_field is None and self.value_field is not None) or (
            self.id_field is not None and self.value_field is None
        ):
            raise ValueError(
                f"Please set both id_field and value_field or none of both. Got id_field = {self.id_field} and value_field = {self.value_field}."
            )

    def has_id_and_value_field(self) -> bool:
        return self.id_field is not None and self.value_field is not None


class JsonlFormat(KeyValueFormatBase):
    """Checks if a given output is a valid jsonl file."""

    def apply_configuration_and_throw_if_invalid(self, configuration: "Optional[dict[str, Any]]"):
        if not configuration:
            configuration = {}

        if CONF_REQUIRED_FIELDS not in configuration:
            configuration[CONF_REQUIRED_FIELDS] = set()

        if CONF_MINIMUM_LINES not in configuration:
            configuration[CONF_MINIMUM_LINES] = 2

        super().apply_configuration_and_throw_if_invalid(configuration)

    def check_format(self, run_output: Path):
        try:
            lines = self.all_lines(run_output)

            if len(lines) < self.minimum_lines:
                return [_fmt.ERROR, f"The *.jsonl file contains only {len(lines)} lines, this is likely an error."]

            return [_fmt.OK, "The jsonl file has the correct format."]
        except Exception as e:
            return [_fmt.ERROR, str(e)]

    def fail_if_json_line_is_not_valid(self, line):
        for field in self.required_fields:
            if not line or field not in line:
                raise ValueError(f'The json line misses the required field "{field}": "{json.dumps(line)}".')
        if self.has_id_and_value_field():
            if not line or self.id_field not in line:
                raise ValueError(f'The json line misses the required field "{self.id_field}": "{json.dumps(line)}".')
            if not line or self.value_field not in line:
                raise ValueError(f'The json line misses the required field "{self.value_field}": "{json.dumps(line)}".')

    def yield_next_entry(self, run_output):
        if (str(run_output).endswith(".jsonl") or str(run_output).endswith(".jsonl.gz")) and run_output.is_file():
            matches = [run_output]
        else:
            matches = [
                run_output / i for i in os.listdir(run_output) if i.endswith(".jsonl") or i.endswith(".jsonl.gz")
            ]

        if len(matches) != 1:
            raise ValueError(
                "No unique *.jsonl file was found, only the files " + str(os.listdir(run_output)) + " were available."
            )

        for line in super().yield_next_entry(matches[0]):
            line = line.strip()
            try:
                parsed_line = json.loads(line)
                self.fail_if_json_line_is_not_valid(parsed_line)
                yield parsed_line
            except json.JSONDecodeError:
                raise ValueError(f'The file {matches[0]} contains a line that could not be parsed: "{line}".')


class ToucheImageRetrieval(JsonlFormat):
    def apply_configuration_and_throw_if_invalid(self, configuration: "Optional[dict[str, Any]]"):
        super().apply_configuration_and_throw_if_invalid(
            {CONF_REQUIRED_FIELDS: ["argument_id", "method", "image_id", "rank", "tag"]}
        )

    def yield_next_entry(self, run_output):
        for i in super().yield_next_entry(run_output):
            yield {
                "qid": i["argument_id"],
                "q0": "0",
                "docno": i["image_id"],
                "rank": i["rank"],
                "score": 1000 - i["rank"],
                "system": i["tag"],
            }


class LongEvalLags(FormatBase):
    def apply_configuration_and_throw_if_invalid(self, configuration: "Optional[dict[str, Any]]"):
        if not configuration or "lags" not in configuration or not configuration["lags"]:
            raise ValueError(
                'Please pass a configuration "lags" that points out on which lags an dataset should run. '
                + 'E.g., {"lags": ["lag-1", "lag-2"]}. '
                + f"I got: {configuration}"
            )

        self.lags = configuration["lags"]
        if isinstance(self.lags, str):
            self.lags = [self.lags]
        format = "run.txt"
        if "format" in configuration:
            format = configuration["format"]
        self.format = format

        self.max_size_mb = 250
        if configuration and hasattr(configuration, "__iter__") and CONF_MAX_SIZE_MB in configuration:
            self.max_size_mb = configuration[CONF_MAX_SIZE_MB]
        self.spot_check = 25000
        if configuration and hasattr(configuration, "__iter__") and CONF_SPOT_CHECK in configuration:
            self.spot_check = configuration[CONF_SPOT_CHECK]

    def check_ir_metadata(self, run_output: Path):
        if not (run_output / "ir-metadata.yml").exists():
            return _fmt.ERROR, "\n\t" + fmt_message(
                f"I expected a file ir-metadata.yml in the directory {run_output} but did not find one.", _fmt.ERROR
            )
        else:
            required_fields = {
                "tag": {"type": str, "default": "ENTER_VALUE_HERE"},
                "actor": {"team": {"type": str, "default": "ENTER_VALUE_HERE"}},
                "research goal": {"description": {"type": str, "default": "ENTER_VALUE_HERE"}},
                "platform": {"software": {"libraries": {"type": list, "default": "ENTER_VALUE_HERE"}}},
                "implementation": {"source": {"repository": {"type": str, "default": "ENTER_VALUE_HERE"}}},
                "data": {"training data": {"name": {"type": str, "default": "ENTER_VALUE_HERE"}}},
                "method": {
                    "automatic": {"type": bool, "default": "ENTER_VALUE_HERE"},
                    "indexing": {
                        "tokenizer": {"type": str, "default": "ENTER_VALUE_HERE"},
                        "stemmer": {"type": str, "default": "ENTER_VALUE_HERE"},
                        "stopwords": {"type": str, "default": "ENTER_VALUE_HERE"},
                    },
                    "retrieval": {
                        "name": {"type": str, "default": "ENTER_VALUE_HERE"},
                        "lexical": {"type": bool, "default": "ENTER_VALUE_HERE"},
                        "deep_neural_model": {"type": bool, "default": "ENTER_VALUE_HERE"},
                        "sparse_neural_model": {"type": bool, "default": "ENTER_VALUE_HERE"},
                        "single_stage_retrieval": {"type": bool, "default": "ENTER_VALUE_HERE"},
                    },
                },
            }
            return check_format(
                run_output / "ir-metadata.yml",
                "ir_metadata",
                {"required_fields": required_fields},
            )

    def check_format(self, run_output: Path):
        ret_msg = f"I will check that the data in {run_output} is valid ..."
        ret_code = _fmt.OK

        if self.check_ir_metadata(run_output)[0] != _fmt.OK:
            return _fmt.ERROR, "\n\t" + fmt_message(
                f"The file {run_output}/ir-metadata.yml is not valid. Errors: " + self.check_ir_metadata(run_output)[1],
                _fmt.ERROR,
            )

        for lag in self.lags:
            status, msg = check_format(
                run_output / lag, self.format, {CONF_MAX_SIZE_MB: self.max_size_mb, CONF_SPOT_CHECK: self.spot_check}
            )
            if status != _fmt.OK:
                ret_code = _fmt.ERROR
                ret_msg += "\n\t" + fmt_message(
                    f"I expected a run file in the subdirectory {lag}. Error: {msg}", _fmt.ERROR
                )
            else:
                ret_msg += "\n\t" + fmt_message(f"The run in subdirectory {lag} is valid.", _fmt.OK)

        ret_msg += "\n\t" + fmt_message("The file ir-metadata.yml is valid.", _fmt.OK)

        return [ret_code, ret_msg]


class PowerAndIdeologyFormat(FormatBase):
    def __init__(self, prefix_pattern, suffix_pattern, languages=None, tasks=None):
        if not languages:
            languages = [
                "at",
                "ba",
                "be",
                "bg",
                "cz",
                "dk",
                "ee",
                "es-ct",
                "es-ga",
                "es-pv",
                "es",
                "fi",
                "fr",
                "gb",
                "gr",
                "hr",
                "hu",
                "is",
                "it",
                "lv",
                "nl",
                "no",
                "pl",
                "pt",
                "rs",
                "se",
                "si",
                "tr",
                "ua",
            ]
        if not tasks:
            tasks = ["orientation", "power", "populism"]

        self.languages = languages
        self.tasks = tasks
        self.prefix_pattern = prefix_pattern
        self.suffix_pattern = suffix_pattern

    def yield_next_entry(self, f: Path):
        for language in self.languages:
            for task in self.tasks:
                matching_files = glob(
                    str(f) + "/" + self.prefix_pattern + task + "-" + language + self.suffix_pattern + ".tsv"
                )
                for file in matching_files:
                    with open(file, "rt") as f:
                        for l in f:
                            if len(l.split("\t")) != 2:
                                raise ValueError(f"Invalid tsv line in {f}. Got '" + l + "'.")
                            l = l.split("\t")
                            yield {"id": l[0], "language": language, task: "task", "label": l[1]}

    def check_format(self, run_output: Path):
        lines = list(self.all_lines(run_output))
        if len(lines) < 3:
            return [
                _fmt.ERROR,
                f"I did not find any files in the power and ideology pattern.\n\n\tThe pattern is:\n\t{self.prefix_pattern}TASK-LANGUAGE{self.suffix_pattern}.tsv",
            ]
        return [_fmt.OK, f"I found {len(lines)} entries in the ideology and power task"]


class GenIrSimulationFormat(JsonlFormat):
    def fail_if_json_line_is_not_valid(self, line):
        if "simulation" not in line:
            raise ValueError('The json line misses the required field "simulation".')
        line = line["simulation"]

        if "configuration" not in line:
            raise ValueError('The json line misses the required field "simulation.configuration".')

        if "userTurns" not in line:
            raise ValueError('The json line misses the required field "simulation.userTurns".')

    def apply_configuration_and_throw_if_invalid(self, configuration: "Optional[dict[str, Any]]"):
        super().apply_configuration_and_throw_if_invalid(configuration)

        required_fields = set()
        minimum_lines = 1

        if configuration and hasattr(configuration, "__iter__"):
            if CONF_REQUIRED_FIELDS in configuration:
                required_fields = set(configuration[CONF_REQUIRED_FIELDS])
            if CONF_MINIMUM_LINES in configuration:
                minimum_lines = int(configuration[CONF_MINIMUM_LINES])

        self.required_fields = required_fields
        self.minimum_lines = minimum_lines


class TsvFormat(KeyValueFormatBase):
    """Checks if a given output is a valid tsv file."""

    def apply_configuration_and_throw_if_invalid(self, configuration: "Optional[dict[str, Any]]"):
        if not configuration:
            configuration = {}

        if CONF_MINIMUM_LINES not in configuration:
            configuration[CONF_MINIMUM_LINES] = 5

        super().apply_configuration_and_throw_if_invalid(configuration)

    def check_format(self, run_output: Path):
        try:
            lines = self.all_lines(run_output)
        except Exception as e:
            return [_fmt.ERROR, str(e)]

        if len(lines) < self.minimum_lines:
            return [_fmt.ERROR, f"The *.tsv file contains only {len(lines)} lines, this is likely an error."]

        if len(lines[0]) < 2:
            return [_fmt.ERROR, "The *.tsv file contains only one column, this is likely an error."]

        return [_fmt.OK, "The tsv file has the correct format."]

    def yield_next_entry(self, run_output):
        matches = [i for i in os.listdir(run_output) if i.endswith(".tsv")]
        if len(matches) != 1:
            msg = "No unique *.tsv file was found, only the files "
            msg += str(os.listdir(run_output)) + " were available."
            raise ValueError(msg)

        f = run_output / matches[0]
        columns = None

        for line in super().yield_next_entry(f):
            l_parsed = line.strip().split("\t")
            if columns is None:
                columns = len(l_parsed)
            if len(l_parsed) != columns:
                raise ValueError("The *.tsv file is invalid: The number of columns varies.")
            yield l_parsed


class MultiAuthorWritingStyleAnalysis(FormatBase):
    def __init__(self, prefix):
        self.prefix = prefix

    def apply_configuration_and_throw_if_invalid(self, configuration: "Optional[dict[str, Any]]"):
        if not configuration:
            configuration = {}

        if CONF_MINIMUM_LINES not in configuration:
            configuration[CONF_MINIMUM_LINES] = 3

        self.minimum_lines = configuration[CONF_MINIMUM_LINES]

    def yield_next_entry(self, run_output):
        suffix = ".json" if self.prefix != "problem" else ".txt"
        matches = (
            glob(f"{run_output}/{self.prefix}-*{suffix}")
            + glob(f"{run_output}/**/{self.prefix}-*{suffix}")
            + glob(f"{run_output}/**/**/{self.prefix}-*{suffix}")
        )

        if len(matches) < self.minimum_lines:
            raise ValueError(
                f"There are no files matching the multi-author-style file pattern of '{self.prefix}-*{suffix}' in the directory {run_output}."
            )

        tasks = set(["easy", "medium", "hard"])

        for f in matches:
            task = [i for i in tasks if i in f]
            if len(task) != 1:
                raise ValueError("fooo")

            with open(f, "r") as fh:
                f = Path(f).relative_to(run_output)
                problem = fh.read()
                if suffix == ".json":
                    ret = {
                        "file": f,
                        "task": task[0],
                    }
                    ret.update(json.loads(problem))
                    yield ret
                else:
                    yield {
                        "file": str(f),
                        "task": task[0],
                        "problem": problem,
                        "paragraphs": problem.split("\n"),
                    }

    def check_format(self, run_output: Path):
        try:
            lines = self.all_lines(run_output)
            return [_fmt.OK, f"Valid Multi-Author-Writing-Style directory with {len(lines)} entries."]
        except Exception as e:
            return [_fmt.ERROR, str(e)]


class TextAlignmentFeaturesFormat(FormatBase):
    """
    The code was refactored from https://github.com/pan-webis-de/pan-code/blob/master/sepln09/pan09-plagiarism-detection-performance-measures.py#L276
    """

    def check_format(self, run_output: Path):
        try:
            lines = self.all_lines(run_output)
        except Exception as e:
            return [_fmt.ERROR, str(e)]

        if len(lines) < 3:
            return [
                _fmt.ERROR,
                f"The text-alignment-feature directory contains only {len(lines)} instances, this is likely an error.",
            ]

        return [_fmt.OK, "The directory has the correct format."]

    def all_lines(self, run_output):
        return self.extract_annotations_from_files(run_output, ["plagiarism", "detected-plagiarism"])

    def extract_annotations_from_file(self, xmlfile, tagnames):
        """Returns a set of plagiarism annotations from an XML file."""
        doc = xml.dom.minidom.parse(xmlfile)
        annotations = []
        if not doc.documentElement.hasAttribute("reference"):
            return annotations
        t_ref = doc.documentElement.getAttribute("reference")
        for node in doc.documentElement.childNodes:
            for tagname in tagnames:
                if (
                    node.nodeType == xml.dom.Node.ELEMENT_NODE
                    and node.hasAttribute("name")
                    and node.getAttribute("name").endswith(tagname)
                ):
                    ann = self.extract_annotation_from_node(node, t_ref)
                    if ann:
                        ann["type"] = tagname
                        annotations.append(ann)
        return annotations

    def extract_annotation_from_node(self, xmlnode, t_ref):
        """Returns a plagiarism annotation from an XML feature tag node."""
        if not (xmlnode.hasAttribute("this_offset") and xmlnode.hasAttribute("this_length")):
            return False
        t_off = int(xmlnode.getAttribute("this_offset"))
        t_len = int(xmlnode.getAttribute("this_length"))
        s_ref, s_off, s_len, ext = "", 0, 0, False
        if (
            xmlnode.hasAttribute("source_reference")
            and xmlnode.hasAttribute("source_offset")
            and xmlnode.hasAttribute("source_length")
        ):
            s_ref = xmlnode.getAttribute("source_reference")
            s_off = int(xmlnode.getAttribute("source_offset"))
            s_len = int(xmlnode.getAttribute("source_length"))
            ext = True
        return {
            "this_reference": t_ref,
            "this_offset": t_off,
            "this_length": t_len,
            "source_reference": s_ref,
            "source_offset": s_off,
            "source_length": s_len,
            "is_external": ext,
        }

    def extract_annotations_from_files(self, path, tagnames):
        """Returns a set of plagiarism annotations from XML files below path."""
        if not os.path.exists(path):
            raise ValueError(f"Path not accessible: {path}")

        annotations = []
        xmlfiles = glob(os.path.join(path, "*.xml"))
        xmlfiles.extend(glob(os.path.join(path, os.path.join("*", "*.xml"))))
        for xmlfile in xmlfiles:
            annotations += self.extract_annotations_from_file(xmlfile, tagnames)
        return annotations


class PanStyleChangeDetectionCorpusFormat(FormatBase):
    """Checks if a given directory contains a valid style-change-detection PAN corpus. An example is located here: https://pan.webis.de/clef24/pan24-web/style-change-detection.html"""

    def check_format(self, run_output: Path):
        try:
            lines = [i for i in self.yield_lines(run_output, False)]
        except Exception as e:
            return [_fmt.ERROR, str(e)]

        if len(lines) < 3:
            return [
                _fmt.ERROR,
                f"The style-change-detection directory contains only {len(lines)} instances, this is likely an error.",
            ]

        return [_fmt.OK, "The directory has the correct format."]

    def all_lines(self, run_output):
        return [i for i in self.yield_lines(run_output, True)]

    def yield_lines(self, directory, load_content):
        matches = (
            glob(f"{directory}/problem-*.txt")
            + glob(f"{directory}/*/problem-*.txt")
            + glob(f"{directory}/*/*/problem-*.txt")
            + glob(f"{directory}/*/*/*/problem-*.txt")
        )

        for i in matches:
            entry = {"id": i.split("/")[-1].replace(".txt", "")}

            if load_content:
                with open(i, "r", newline="") as f:
                    entry["text"] = f.read()

            yield entry


class PanStyleChangeDetectionPredictionFormat(FormatBase):
    """Checks if a given directory contains a valid style-change-detection PAN predictions. An example is located here: https://pan.webis.de/clef24/pan24-web/style-change-detection.html"""

    def check_format(self, run_output: Path):
        try:
            lines = [i for i in self.yield_lines(run_output, False)]
        except Exception as e:
            return [_fmt.ERROR, str(e)]

        if len(lines) < 3:
            return [
                _fmt.ERROR,
                f"The style-change-detection directory contains only {len(lines)} instances, this is likely an error.",
            ]

        return [_fmt.OK, "The directory has the correct format."]

    def all_lines(self, run_output):
        return [i for i in self.yield_lines(run_output, True)]

    def yield_lines(self, directory, load_content):
        matches = (
            glob(f"{directory}/*-problem-*.json")
            + glob(f"{directory}/*/*-problem-*.json")
            + glob(f"{directory}/*/*/*-problem-*.json")
            + glob(f"{directory}/*/*/*/*-problem-*.json")
        )

        for i in matches:
            prediction_type, problem_id = i.split("/")[-1].replace(".json", "").split("-problem-")
            entry = {"problem": problem_id, "type": prediction_type}

            if load_content:
                with open(i, "r", newline="") as f:
                    entry.update(json.loads(f.read()))

            yield entry


class TextAlignmentCorpusFormat(FormatBase):
    """Checks if a given directory contains a valid PAN text alignment corpus. The code was refactored from https://github.com/pan-webis-de/pan-code/blob/master/clef12/text-alignment/pan12-text-alignment-baseline.py#L202"""

    def check_format(self, run_output: Path):
        try:
            lines = [i for i in self.yield_lines(run_output, False)]
        except Exception as e:
            return [_fmt.ERROR, str(e)]

        if len(lines) < 3:
            return [
                _fmt.ERROR,
                f"The text-alignment directory contains only {len(lines)} instances, this is likely an error.",
            ]

        return [_fmt.OK, "The directory has the correct format."]

    def all_lines(self, run_output):
        return [i for i in self.yield_lines(run_output, True)]

    def yield_lines(self, run_output, load_content):
        matches = [i for i in os.listdir(run_output) if i.endswith("pairs")]
        if len(matches) != 1:
            msg = "No unique pair file was found, only the files "
            msg += str(os.listdir(run_output)) + " were available."
            raise ValueError(msg)

        f = run_output / matches[0]
        with open(f, "r") as tsv_file:
            for l in tsv_file:
                l_parsed = l.strip().split()
                if len(l_parsed) != 2:
                    raise ValueError("The pair file is invalid: I expect always two lines per entry.")
                entry = {}
                for k, v in zip(l_parsed, ["suspicious", "source"]):
                    entry[f"{v}_document_id"] = k.replace(".txt", "")
                    matches = (
                        glob(f"{run_output}/{k}")
                        + glob(f"{run_output}/{k}.txt")
                        + glob(f"{run_output}/*/{k}")
                        + glob(f"{run_output}/*/{k}.txt")
                    )

                    if len(matches) != 1:
                        msg = f"No unique text file was found for id {k}. Only "
                        msg += str(os.listdir(run_output)) + " were available."

                        raise ValueError(msg)
                    if load_content:
                        with open(matches[0]) as f:
                            entry[f"{v}_document_text"] = f.read()

                yield entry


class QueryProcessorFormat(JsonlFormat):
    """Checks if a given output is a valid query processor output in JSONL format."""

    def __init__(self, qid_name="qid"):
        super().__init__()
        self.qid_name = qid_name

    def apply_configuration_and_throw_if_invalid(self, configuration):
        self.minimum_lines = 1

    def fail_if_json_line_is_not_valid(self, line):
        if "qid" not in line and "query_id" not in line:
            raise ValueError('At least one of "qid" or "query_id" fields is required.')

        if "qid" in line and "query_id" in line and line["qid"] != line["query_id"]:
            raise ValueError('If both "qid" and "query_id" fields are present, they must be equal.')

    def yield_next_entry(self, run_output):
        seen_query_ids = set()

        for line in super().yield_next_entry(run_output):
            # Create a new dictionary to avoid modifying the original
            normalized_line = dict(line)

            if "qid" in line:
                query_id_value = line["qid"]
            elif "query_id" in line:
                query_id_value = line["query_id"]

            query_id_value = str(query_id_value)

            if query_id_value in seen_query_ids:
                raise ValueError(f"Query ID {query_id_value} is not unique.")
            seen_query_ids.add(query_id_value)

            # Remove redundant ID fields if they differ from the normalized name
            if "qid" in normalized_line and "qid" != self.qid_name:
                del normalized_line["qid"]
            if "query_id" in normalized_line and "query_id" != self.qid_name:
                del normalized_line["query_id"]

            normalized_line[self.qid_name] = query_id_value

            yield normalized_line


class DocumentProcessorFormat(JsonlFormat):
    """Checks if a given output is a valid document processor output in JSONL format."""

    def __init__(self, docno_name="docno"):
        super().__init__()
        self.docno_name = docno_name

    def apply_configuration_and_throw_if_invalid(self, configuration):
        super().apply_configuration_and_throw_if_invalid(configuration)
        self.minimum_lines = 1
        self.doc_ids = ["docno", "docid", "doc_id"]

    def extract_docno_or_fail(self, line):
        ret = None
        for doc_id in self.doc_ids:
            if doc_id in line and line[doc_id]:
                if ret is None or ret == line[doc_id]:
                    ret = line[doc_id]
                else:
                    raise ValueError(f"Inconsistent fields for the document identifier: {line}.")

        if ret is None:
            raise ValueError(f"At least one of {self.doc_ids} fields is required.")
        return ret

    def fail_if_json_line_is_not_valid(self, line):
        self.extract_docno_or_fail(line)

    def yield_next_entry(self, run_output):
        seen_query_ids = set()

        for line in super().yield_next_entry(run_output):
            # Create a new dictionary to avoid modifying the original
            normalized_line = dict(line)
            id_value = self.extract_docno_or_fail(normalized_line)
            for doc_id in self.doc_ids:
                if doc_id in normalized_line:
                    del normalized_line[doc_id]
            normalized_line[self.docno_name] = id_value
            yield normalized_line


class IrDatasetsCorpus(FormatBase):
    def __init__(self):
        self.docs_processor = DocumentProcessorFormat()
        self.query_processor = QueryProcessorFormat()

    def apply_configuration_and_throw_if_invalid(self, configuration):
        self.docs_processor.apply_configuration_and_throw_if_invalid(configuration)
        self.query_processor.apply_configuration_and_throw_if_invalid(configuration)

    def yield_next_entry(self, run_output):
        raise ValueError("not implemented")

    def check_format(self, run_output: Path):
        code, message = self.query_processor.check_format(Path(run_output) / "queries.jsonl")

        if code != _fmt.OK:
            return [_fmt.ERROR, f"No ir-dataset found. {message}"]

        code, message = self.docs_processor.check_format(Path(run_output) / "corpus.jsonl.gz")

        if code != _fmt.OK:
            return [_fmt.ERROR, f"No ir-dataset found. {message}"]

        return [_fmt.OK, "Valid ir-dataset found."]


class LightningIrDocumentEmbeddings(FormatBase):
    def yield_next_entry(self, run_output):
        raise ValueError("not implemented")

    def check_format(self, run_output: Path):
        for expected_file in ["doc-ids.txt", "doc-embeddings.npz"]:
            if (
                len(glob(f"{Path(run_output)}/{expected_file}")) == 0
                and len(glob(f"{Path(run_output)}/**/{expected_file}")) == 0
            ):
                return [_fmt.ERROR, f"No lightning-ir embeddings found. I expected a file {expected_file}"]
        return [_fmt.OK, "Valid lightning-ir embeddings found."]


class LightningIrQueryEmbeddings(FormatBase):
    def yield_next_entry(self, run_output):
        raise ValueError("not implemented")

    def check_format(self, run_output: Path):
        for expected_file in ["query-embeddings.npz", "query-ids.txt"]:
            if (
                len(glob(f"{Path(run_output)}/{expected_file}")) == 0
                and len(glob(f"{Path(run_output)}/**/{expected_file}")) == 0
            ):
                return [_fmt.ERROR, f"No lightning-ir embeddings found. I expected a file {expected_file}"]
        return [_fmt.OK, "Valid lightning-ir embeddings found."]


class TerrierIndex(FormatBase):
    def apply_configuration_and_throw_if_invalid(self, configuration):
        pass

    def check_format(self, run_output: Path):
        if not run_output.is_dir():
            return [_fmt.ERROR, "The directory is no valid terrier index, it misses an index directory."]

        if not (run_output / "index").is_dir():
            return [_fmt.ERROR, "The directory is no valid terrier index, it misses an index directory."]

        if not (run_output / "index" / "data.properties").is_file():
            return [_fmt.ERROR, "The directory is no valid terrier index, it misses an index/data.properties file."]

        if not (run_output / "index" / "data.meta.idx").is_file():
            return [_fmt.ERROR, "The directory is no valid terrier index, it misses an index/data.meta.idx file."]

        return [_fmt.OK, "The directory seems to be a valid pyterrier index."]

    def yield_next_entry(self, f):
        pass


class IrMetadataFormat(FormatBase):
    """Checks if a given output contains valid ir_metadata."""

    def __init__(self, positive_fields=("platform", "implementation", "resources")):
        self.positive_fields = positive_fields

    def apply_configuration_and_throw_if_invalid(self, configuration):
        self.required_fields = None
        if configuration and hasattr(configuration, "__iter__") and CONF_REQUIRED_FIELDS in configuration:
            self.required_fields = configuration[CONF_REQUIRED_FIELDS]

    def check_format(self, run_output: Path):
        try:
            lines = list(self.yield_lines(run_output))

            if self.required_fields:
                for line in lines:
                    errors = self.report_missing_fields(line["content"], self.required_fields)
                    msg = ""
                    if len(errors) > 0:
                        for error in errors:
                            msg += f"\n\t\t- {error}"
                        return [_fmt.ERROR, msg]

            if len(lines) < 1:
                return [_fmt.ERROR, "At least one valid ir_metadata file was expected, but there was none."]

            return [_fmt.OK, "The output provides valid ir_metadata."]
        except Exception as e:
            return [_fmt.ERROR, str(e)]

    def all_lines(self, run_output):
        return [i for i in self.yield_lines(run_output)]

    def report_missing_fields(self, yml_dict, required_fields, prefix=None):
        ret = []
        for k, v in required_fields.items():
            k_display = ("" if prefix is None else prefix + ".") + k
            try:
                if not yml_dict or k not in yml_dict:
                    ret.append(f"The required field {k_display} is missing.")

                    continue
            except Exception as e:
                ret.append(f"The required field {k_display} is missing.")
                continue

            check_type = "type" in v and "default" in v
            values = [yml_dict[k]] if not isinstance(yml_dict[k], list) or check_type else yml_dict[k]

            for val in values:
                if check_type:
                    if val == v["default"] or (isinstance(val, str) and v["default"] in val):
                        ret.append(f"The required field {k_display} still contains the default value {v['default']}.")
                    elif not isinstance(val, v["type"]):
                        ret.append(
                            f"The required field {k_display} has the wrong type {type(val)}. I expected {v['type']}."
                        )
                else:
                    ret.extend(self.report_missing_fields(val, v, k_display))

        return ret

    def yield_lines(self, run_output: Path):
        import yaml

        candidates = [str(run_output)]
        for pattern in ["/*.yml", "/*.yaml", "/.*.yml", "/.*.yaml"]:
            for depth in ["", "/**", "/**/**", "/.**", "/.**/**", "/.**/.**"]:
                candidates += glob(str(run_output) + depth + pattern)

        for candidate in candidates:
            if not Path(candidate).exists() or not Path(candidate).is_file():
                continue

            with open(candidate) as stream:
                try:
                    content = yaml.safe_load(stream)
                    if not content:
                        raise ValueError("The ir_metadata file is empty.")

                    if not self.required_fields:
                        at_least_one_positive_field = any(i in content for i in self.positive_fields)

                        if not at_least_one_positive_field:
                            continue
                    name = candidate.split("/")[-1]
                    if len(candidate.split("/")) > 1 and candidate.split("/")[-2].startswith("."):
                        name = str(candidate.split("/")[-2]) + "/" + name

                    yield {"name": name, "content": content}
                except yaml.YAMLError:
                    pass


FORMAT_TO_CHECK = {
    "run.txt": RunFormat,
    "*.jsonl": JsonlFormat,
    "*.tsv": TsvFormat,
    "text-alignment-corpus": TextAlignmentCorpusFormat,
    "text-alignment-features": TextAlignmentFeaturesFormat,
    "style-change-detection-corpus": PanStyleChangeDetectionCorpusFormat,
    "style-change-detection-predictions": PanStyleChangeDetectionPredictionFormat,
    "GenIR-Simulation": GenIrSimulationFormat,
    "query-processor": QueryProcessorFormat,
    "document-processor": DocumentProcessorFormat,
    "ir_metadata": IrMetadataFormat,
    "lsr-benchmark-inputs": LearnedSparseRetrievalInputs,
    "qrels.txt": QrelFormat,
    "LongEvalLags": LongEvalLags,
    "terrier-index": TerrierIndex,
    "multi-author-writing-style-analysis-problems": lambda: MultiAuthorWritingStyleAnalysis("problem"),
    "multi-author-writing-style-analysis-solutions": lambda: MultiAuthorWritingStyleAnalysis("solution-problem"),
    "multi-author-writing-style-analysis-truths": lambda: MultiAuthorWritingStyleAnalysis("truth-problem"),
    "power-and-identification-truths": lambda: PowerAndIdeologyFormat("", "-*-labels"),
    "power-and-identification-predictions": lambda: PowerAndIdeologyFormat("*", "*-predictions"),
    "touche-image-retrieval": ToucheImageRetrieval,
    "ir-dataset-corpus": IrDatasetsCorpus,
    "lightning-ir-document-embeddings": LightningIrDocumentEmbeddings,
    "lightning-ir-query-embeddings": LightningIrQueryEmbeddings,
}

SUPPORTED_FORMATS = set(sorted(list(FORMAT_TO_CHECK.keys())))


def check_format_configuration_if_valid(
    format: Union[str, Sequence[str]], configuration: "Optional[dict[str, Any]]" = None
) -> "FormatBase":
    if isinstance(format, str):
        ret = FORMAT_TO_CHECK[format]()
        ret.apply_configuration_and_throw_if_invalid(configuration)
        return ret
    else:
        ret = []
        for i in format:
            ret.append(check_format_configuration_if_valid(i, configuration))

        if len(ret) == 1:
            return ret[0]
        else:
            return ret


def lines_if_valid(
    run_output: Path, format: "Union[str, Sequence[str]]", configuration: "Optional[dict[str, Any]]" = None
):
    """Load all lines from a user file if they are valid

    Args:
        format (Union[str, Sequence[str]]): The allowed format or a list of allowed formats.
        run_output (Path): the output produced by some run that is to-be checked.
        configuration (Optional[dict[str, Any]]): the configuration to apply to the formatter.
    """
    if not isinstance(format, str) and isinstance(format, Iterable):
        if len(format) == 0 or len(format) > 1:
            raise ValueError("Configuration error, I do not know in which format to read the file, I got {format}")
        else:
            format = format[0]

    if format not in SUPPORTED_FORMATS:
        raise ValueError(f"Format {format} is not supported. Supported formats are {SUPPORTED_FORMATS}.")

    checker = check_format_configuration_if_valid(format, configuration)
    result, msg = checker.check_format(run_output)
    if result != _fmt.OK:
        raise ValueError(msg)

    return checker.all_lines(run_output)


def report_valid_formats(run_output: Path) -> Dict[str, Any]:
    valid_formats: Dict[str, Any] = {}
    if _fmt.OK == check_format(run_output, "ir_metadata")[0]:
        valid_formats["ir_metadata"] = sorted([i["name"] for i in lines_if_valid(run_output, "ir_metadata")])

    for f in ["run.txt", "query-processor", "document-processor", "terrier-index"]:
        if _fmt.OK == check_format(run_output, f)[0]:
            valid_formats[f] = True

    return valid_formats


def check_format(
    run_output: Path, format: "Union[str, Sequence[str]]", configuration: "Optional[dict[str, Any]]" = None
):
    """Check if the provided run output is in the specified format. Provides debug messages intended for users.

    Args:
        format (Union[str, Sequence[str]]): The allowed format or a list of allowed formats.
        run_output (Path): the output produced by some run that is to-be checked.
        configuration (Optional[dict[str, Any]]): the configuration to apply to the formatter.
    """
    if not isinstance(format, str) and isinstance(format, Iterable):
        ret = {}
        for f in format:
            ret[f] = check_format(run_output, f, configuration)

        if all(i[0] == _fmt.OK for i in ret.values()):
            return [_fmt.OK, "The output is valid."]

        error_msg = [i[1] for i in ret.values() if i[0] != _fmt.OK]

        if "ir_metadata" in ret:
            metadata_valid = ret["ir_metadata"][0] == _fmt.OK
            one_payload_valid = any(v[0] == _fmt.OK for k, v in ret.items() if k != "ir_metadata")

            if metadata_valid and one_payload_valid:
                return [_fmt.OK, "The output is valid."]
            else:
                return [_fmt.ERROR, "The output is not valid. Problems: " + " ".join(error_msg)]

        ret = {k: v for k, v in ret.items() if v[0] == _fmt.OK}

        if len(ret) > 0:
            return [_fmt.OK, list(ret.values())[0][1]]
        else:
            return [_fmt.ERROR, "The output is not valid. Problems: " + " ".join(error_msg)]

    if format not in SUPPORTED_FORMATS:
        raise ValueError(f"Format {format} is not supported. Supported formats are {SUPPORTED_FORMATS}.")

    checker = check_format_configuration_if_valid(format, configuration)

    return checker.check_format(run_output)



import unittest

import pandas as pd

from tira.local_client import Client

# .. todo:: this file still uses "unclean" assertions and should be converted to use unittest assertions


def keyphrase_extraction(docs):
    queries = pd.DataFrame([{"docno": str(i)} for i in docs])
    tira = Client("tests/resources/")
    query_segmentation = tira.pt.transform_documents(
        "ir-benchmarks/webis-keyphrase-extraction/BCExtractorFO", dataset="d1"
    )

    ret = query_segmentation(queries)
    return {i["docno"]: i["keyphrases"] for _, i in ret.iterrows()}


def doc_processor(docs, name):
    queries = pd.DataFrame([{"docno": str(i)} for i in docs])
    tira = Client("tests/resources/")
    query_segmentation = tira.pt.transform_documents("ir-benchmarks/tira-ir-starters/" + name, dataset="d1")

    ret = query_segmentation(queries)
    return {i["docno"]: i["key"] for _, i in ret.iterrows()}


class TestPtDocumentProcessingLoaderTest(unittest.TestCase):
    def test_for_loading_keyphrase_extraction_for_multiple_documents(self):
        expected = {
            "FT921-3964": [
                "increased economic aid",
                "king hussein",
                "increase aid",
                "president george bush",
                "tariq aziz iraq's deputy prime minister",
            ],
            "LA061489-0137": [
                "public appearances gorbachev",
                "crowds",
                "government ministries gorbachev",
                "place gorbachev",
                "west german well-wishers",
            ],
        }

        actual = keyphrase_extraction(["FT921-3964", "LA061489-0137"])

        assert len(actual) == 2
        assert actual["FT921-3964"] == expected["FT921-3964"]
        assert actual["LA061489-0137"] == expected["LA061489-0137"]

    def test_for_loading_keyphrase_extraction_for_single_documents(self):
        expected = {
            "LA120390-0047": [
                "even friendly technology takeovers",
                "largest technology takeover",
                "computer industry takeover",
                "statement sunday",
                "attempts",
            ]
        }

        actual = keyphrase_extraction(["LA120390-0047"])

        assert len(actual) == 1
        assert actual["LA120390-0047"] == expected["LA120390-0047"]

    def test_pyterrier_can_be_loaded(self):
        from tira.third_party_integrations import ensure_pyterrier_is_loaded

        ensure_pyterrier_is_loaded()

    def test_document_transformation_with_docno(self):
        expected = {"doc-01": "value"}
        actual = doc_processor(["doc-01"], "tiny-example-01")

        self.assertEqual(expected, actual)

    def test_document_transformation_with_doc_id(self):
        expected = {"doc-01": "value"}
        actual = doc_processor(["doc-01"], "tiny-example-02")

        self.assertEqual(expected, actual)

    def test_document_features_with_doc_id(self):
        expected = ["value"]

        documents = pd.DataFrame([{"docno": str(i)} for i in ["doc-01"]])
        tira = Client("tests/resources/")
        doc_features = tira.pt.doc_features("ir-benchmarks/tira-ir-starters/tiny-example-02", dataset="d1")

        actual = doc_features(documents)

        self.assertEqual(1, len(actual))
        self.assertEqual(expected, actual.iloc[0].to_dict()["features"])

    def test_document_features_with_docno(self):
        expected = ["value"]

        documents = pd.DataFrame([{"docno": str(i)} for i in ["doc-01"]])
        tira = Client("tests/resources/")
        doc_features = tira.pt.doc_features("ir-benchmarks/tira-ir-starters/tiny-example-01", dataset="d1")

        actual = doc_features(documents)

        self.assertEqual(1, len(actual))
        self.assertEqual(expected, actual.iloc[0].to_dict()["features"])



import unittest

from tira.rest_api_client import Client

# .. todo:: this file still uses "unclean" assertions and should be converted to use unittest assertions


class TestRestAPI(unittest.TestCase):
    tira: Client

    @classmethod
    def setUpClass(cls):
        TestRestAPI.tira = Client()

    def test_all_softwares_works_for_tirex(self):
        actual = TestRestAPI.tira.all_softwares("ir-benchmarks")

        assert "ir-benchmarks/tira-ir-starter/ChatNoir" in actual
        assert "ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)" in actual

    def test_details_of_public_software_01(self):
        actual = TestRestAPI.tira.docker_software("ir-benchmarks/tira-ir-starter/TF_IDF (tira-ir-starter-pyterrier)")

        assert actual is not None
        assert actual["ir_re_ranker"] is False

    def test_output_of_run(self):
        actual = TestRestAPI.tira.get_run_output(
            "ir-benchmarks/tira-ir-starter/BM25 Re-Rank (tira-ir-starter-pyterrier)", "antique-test-20230107-training"
        )
        actual = open(actual / "run.txt", "r").read()

        assert actual.startswith("8293 Q0 8293_1 1 33.568841093129976 pyterrier.default_pipelines.wmodel_text_scorer")

    def test_output_of_run_for_redudant_dataset_id(self):
        actual = TestRestAPI.tira.get_run_output(
            "ir-benchmarks/tira-ir-starter/BM25 Re-Rank (tira-ir-starter-pyterrier)",
            "ir-benchmarks/antique-test-20230107-training",
        )
        actual = open(actual / "run.txt", "r").read()

        assert actual.startswith("8293 Q0 8293_1 1 33.568841093129976 pyterrier.default_pipelines.wmodel_text_scorer")

    def test_load_dataset_01(self):
        actual = TestRestAPI.tira.download_dataset("workshop-on-open-web-search", "retrieval-20231027-training")
        actual = open(actual / "queries.jsonl", "r").read()

        assert actual.startswith('{"qid": "1", "query": "hubble telescope achievements"}')

    def test_load_dataset_02(self):
        actual = TestRestAPI.tira.download_dataset(
            "workshop-on-open-web-search", "workshop-on-open-web-search/retrieval-20231027-training"
        )
        actual = open(actual / "queries.jsonl", "r").read()

        assert actual.startswith('{"qid": "1", "query": "hubble telescope achievements"}')



import unittest

from tira.check_format import check_format

from . import (
    _ERROR,
    _OK,
    EMPTY_OUTPUT,
    TSV_OUTPUT_VALID,
    VALID_QREL_PATH,
)


class TestCheckJsonlForNonExistingFormats(unittest.TestCase):
    def test_invalid_validator_on_empty_output(self):
        expected = [_ERROR, "No unique qrels.txt file was found, only the files ['.gitkeep'] were available."]
        actual = check_format(EMPTY_OUTPUT, "qrels.txt")
        self.assertEqual(expected, actual)

    def test_invalid_tsv(self):
        expected = [_ERROR, "No unique qrels.txt file was found, only the files ['predictions.tsv'] were available."]
        actual = check_format(TSV_OUTPUT_VALID, "qrels.txt")
        self.assertEqual(expected, actual)

    def test_valid_qrel_file(self):
        expected = [_OK, "The qrels are valid."]
        actual = check_format(VALID_QREL_PATH, "qrels.txt")
        self.assertEqual(expected, actual)



import unittest

from tira.check_format import check_format

from . import _ERROR, _OK, EMPTY_OUTPUT, LIGTHNING_IR_QUERY_EMBEDDINGS, TSV_OUTPUT_VALID, VALID_QREL_PATH


class TestLightningIrQueryEmbeddingFormat(unittest.TestCase):
    def test_invalid_validator_on_empty_output(self):
        actual = check_format(EMPTY_OUTPUT, "lightning-ir-query-embeddings")
        self.assertEqual(_ERROR, actual[0])
        self.assertIn("No lightning-ir embeddings found.", actual[1])

    def test_invalid_tsv(self):
        actual = check_format(TSV_OUTPUT_VALID, "lightning-ir-query-embeddings")
        self.assertEqual(_ERROR, actual[0])
        self.assertIn("No lightning-ir embeddings found.", actual[1])

    def test_invalid_qrel_file(self):
        actual = check_format(VALID_QREL_PATH, "lightning-ir-query-embeddings")
        self.assertEqual(_ERROR, actual[0])
        self.assertIn("No lightning-ir embeddings found.", actual[1])

    def test_valid_query_embeddings(self):
        actual = check_format(LIGTHNING_IR_QUERY_EMBEDDINGS, "lightning-ir-query-embeddings")
        self.assertEqual(_OK, actual[0])
        self.assertIn("Valid lightning-ir embeddings found", actual[1])

    def test_valid_query_embeddings_recursive(self):
        actual = check_format(LIGTHNING_IR_QUERY_EMBEDDINGS.parent, "lightning-ir-query-embeddings")
        self.assertEqual(_OK, actual[0])
        self.assertIn("Valid lightning-ir embeddings found", actual[1])



import tempfile
import unittest

from tira.local_execution_integration import LocalExecutionIntegration as Client

# .. todo:: there are no assertions in this file


class TestDockerIntegration(unittest.TestCase):
    def test_export_of_file_from_bash_image(self):
        tira = Client()
        local_file = tempfile.NamedTemporaryFile().name
        tira.export_file_from_software("/etc/issue", local_file, image="bash:alpine3.16")

        actual = open(local_file).read()
        expected = """Welcome to Alpine Linux 3.16
Kernel \\r on an \\m (\\l)

"""
        assert actual == expected

    def test_export_of_file(self):
        tira = Client()
        local_file = tempfile.NamedTemporaryFile().name
        tira.export_file_from_software("/etc/alpine-release", local_file, image="bash:alpine3.16")

        actual = open(local_file).read()
        expected = """3.16.5\n"""
        assert actual == expected

    def test_extraction_of_entrypoint(self):
        expected = "/docker-entrypoint.sh"

        tira = Client()
        actual = tira.extract_entrypoint(image="docker.io/bash:alpine3.16")

        assert actual == expected

    # def test_execution_of_software(self):
    #    tira = Client()
    #    tira.run(image='bash:alpine3.16', command='sleep 2s;', input_dir="/tmp/input", output_dir="/tmp/output")



import unittest
from pathlib import Path

from tira.tira_client import TiraClient

RESOURCE_DIR = Path(__file__).parent / "resources"


def submit_task(directory, dataset="train"):
    client = TiraClient()
    return client.submit_dataset(directory, "task", dataset, True, skip_baseline=True)


class TestTaskSubmissions(unittest.TestCase):
    def test_fails_for_non_existing_directory(self):
        actual = submit_task(RESOURCE_DIR / "does-not-exist")
        self.assertIsNone(actual)

    def test_works_for_valid_directory(self):
        actual = submit_task(RESOURCE_DIR / "example-datasets" / "multi-author-analysis")
        self.assertIsNotNone(actual)

    def test_fails_for_non_existing_dataset(self):
        actual = submit_task(RESOURCE_DIR / "example-datasets" / "multi-author-analysis", "does-not-exist")
        self.assertIsNone(actual)

    def test_works_for_lsr_benchmark(self):
        actual = submit_task(RESOURCE_DIR / "example-datasets" / "learned-sparse-retrieval")
        self.assertIsNotNone(actual)



import unittest
from pathlib import Path

from . import (
    _ERROR,
    _OK,
    EMPTY_OUTPUT,
    IR_METADATA_INVALID,
    IR_METADATA_SINGLE_VALID,
    VALID_RUN_OUTPUT,
    VALID_RUN_WITH_METADATA_OUTPUT,
)


def check_format(run_output: Path):
    from tira.check_format import check_format as cf

    return cf(run_output, ["run.txt", "ir_metadata"])


class TestMultipleFormats(unittest.TestCase):
    def test_error_for_empty_dir(self):
        expected = [_ERROR, "The output is not valid. Problems: No file run.txt or run.txt.gz was"]
        actual = check_format(EMPTY_OUTPUT)
        self.assertEqual(expected[0], actual[0])
        self.assertIn(expected[1], actual[1])

    def test_error_for_ir_metadata_invalid(self):
        expected = [_ERROR, "The output is not valid. Problems: No file run.txt or run.txt.gz was"]
        actual = check_format(IR_METADATA_INVALID)
        self.assertEqual(expected[0], actual[0])
        self.assertIn(expected[1], actual[1])

    def test_error_for_run_invalid(self):
        expected = [_ERROR, "The output is not valid. Problems: No file run.txt or run.txt.gz was"]
        actual = check_format(IR_METADATA_SINGLE_VALID)
        self.assertEqual(expected[0], actual[0])
        self.assertIn(expected[1], actual[1])

    def test_error_for_metadata_invalid(self):
        expected = [_ERROR, "The output is not valid. Problems: At least one valid ir_metadata file was expected"]
        actual = check_format(VALID_RUN_OUTPUT)
        self.assertEqual(expected[0], actual[0])
        self.assertIn(expected[1], actual[1])

    def test_valid_input(self):
        expected = [_OK, "The output is valid"]
        actual = check_format(VALID_RUN_WITH_METADATA_OUTPUT)
        self.assertEqual(expected[0], actual[0])
        self.assertIn(expected[1], actual[1])



import os
import unittest

import ir_datasets

from tira.third_party_integrations import ensure_pyterrier_is_loaded, load_rerank_data


class TestLoadRerankData(unittest.TestCase):
    def test_loading_rerank_data_from_local_file(self):
        re_rank_data = load_rerank_data(default="tests/resources/re-ranking-outputs/")

        self.assertEqual(3, len(re_rank_data))
        self.assertEqual(
            {"qid": "1", "query": "query 1", "docno": "doc-1", "text": "Text of doc-1", "rank": 1, "score": 10},
            re_rank_data.iloc[0].to_dict(),
        )
        self.assertEqual(
            {"qid": "3", "query": "query 3", "docno": "doc-3", "text": "Text of doc-3", "rank": 1, "score": 1},
            re_rank_data.iloc[2].to_dict(),
        )

    def test_loading_rerank_data_from_remote_file_01(self):
        re_rank_data = load_rerank_data(default="workshop-on-open-web-search/re-ranking-20231027-training")

        self.assertEqual(6, len(re_rank_data))
        self.assertEqual(
            {
                "qid": "2",
                "query": "how to exit vim?",
                "docno": "doc-1",
                "text": (
                    "Press ESC key, then the : (colon), and type the wq command after the colon and hit the Enter key"
                    " to save and leave Vim."
                ),
                "rank": 1,
                "score": 10,
            },
            re_rank_data.iloc[2].to_dict(),
        )

    def test_loading_rerank_data_from_remote_file_02(self):
        re_rank_data = load_rerank_data(default="workshop-on-open-web-search/re-ranking-20231027-training")

        self.assertEqual(6, len(re_rank_data))
        self.assertEqual(
            {
                "qid": "3",
                "query": "signs heart attack",
                "docno": "doc-5",
                "text": (
                    "Common heart attack symptoms include: (1) Chest pain, (2) Pain or discomfort that spreads to the"
                    " shoulder, arm, back, neck, jaw, teeth or sometimes the upper belly, etc."
                ),
                "rank": 1,
                "score": 10,
            },
            re_rank_data.iloc[4].to_dict(),
        )

    def test_loading_of_scored_docs_from_ir_datasets(self):
        from tira.third_party_integrations import ir_datasets

        dataset = ir_datasets.load("workshop-on-open-web-search/re-ranking-20231027-training")
        actual = []

        for i in dataset.scoreddocs_iter():
            actual.append(i)

        self.assertEqual(6, len(actual))
        self.assertEqual("3", actual[4].query_id)
        self.assertEqual(10, actual[4].score)
        self.assertEqual("doc-5", actual[4].doc_id)
        self.assertEqual("signs heart attack", actual[4].query.default_text())
        self.assertEqual(
            "Common heart attack symptoms include: (1) Chest pain, (2) Pain or discomfort that spreads to the shoulder,"
            " arm, back, neck, jaw, teeth or sometimes the upper belly, etc.",
            actual[4].document.default_text(),
        )

    def test_loading_of_scored_docs_from_ir_datasets_within_tira(self):
        os.environ["TIRA_INPUT_DATASET"] = "tests/resources/re-ranking-outputs"
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        dataset = ir_datasets.load("workshop-on-open-web-search/re-ranking-20231027-training")
        print(list(dataset.scoreddocs))
        del os.environ["TIRA_INPUT_DATASET"]
        self.assertEqual(len(list(dataset.scoreddocs)), 3)
        self.assertEqual("doc-1", list(dataset.scoreddocs)[0].doc_id)
        self.assertEqual("1", list(dataset.scoreddocs)[0].query_id)
        self.assertEqual(10, list(dataset.scoreddocs)[0].score)
        self.assertEqual("Text of doc-1", list(dataset.scoreddocs)[0].document.default_text())
        self.assertEqual("query 1", list(dataset.scoreddocs)[0].query.default_text())



import os
import unittest

import ir_datasets

from tira.third_party_integrations import (
    ensure_pyterrier_is_loaded,
    load_ir_datasets,
    register_rerank_data_to_ir_datasets,
)

# .. todo:: this file still uses "unclean" assertions and should be converted to use unittest assertions


class TestIRDatasets(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        register_rerank_data_to_ir_datasets("tests/sample-tirex-rerank-data", "tirex-sample-dataset", "vaswani")

    def test_for_existing_document_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        dataset = ir_datasets.load("tirex-sample-dataset")

        expected_text = (
            "microwave spectroscopy  includes chapters on spectroscope technique\n"
            "and design on measurements on gases liquids and solids on nuclear\n"
            "properties on molecular structure and on further possible applications\n"
            "of microwaves\n"
        )

        assert expected_text == dataset.docs_store().get("8172").text
        assert "8172" == dataset.docs_store().get("8172").doc_id

    def test_for_existing_document_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        dataset = ir_datasets.load("tirex-sample-dataset")

        expected_text = (
            "microwave spectroscopy  includes chapters on spectroscope technique\n"
            "and design on measurements on gases liquids and solids on nuclear\n"
            "properties on molecular structure and on further possible applications\n"
            "of microwaves\n"
        )

        assert expected_text == dataset.docs_store().get("8172").text
        assert "8172" == dataset.docs_store().get("8172").doc_id

    def test_for_non_existing_document_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        dataset = ir_datasets.load("tirex-sample-dataset")

        assert dataset.docs_store().get("does-not-exist") is None

    def test_for_non_existing_document_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        dataset = ir_datasets.load("tirex-sample-dataset")

        assert dataset.docs_store().get("does-not-exist") is None

    def test_for_queries_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        dataset = ir_datasets.load("tirex-sample-dataset")

        assert len(list(dataset.queries_iter())) == 1
        assert (
            "MEASUREMENT OF DIELECTRIC CONSTANT OF LIQUIDS BY THE USE OF MICROWAVE TECHNIQUES\n"
            == list(dataset.queries_iter())[0].text
        )

    def test_for_queries_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        dataset = ir_datasets.load("tirex-sample-dataset")

        assert len(list(dataset.queries_iter())) == 1
        assert (
            "MEASUREMENT OF DIELECTRIC CONSTANT OF LIQUIDS BY THE USE OF MICROWAVE TECHNIQUES\n"
            == list(dataset.queries_iter())[0].text
        )

    def test_for_scoreddocs_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        dataset = ir_datasets.load("tirex-sample-dataset")
        print(list(dataset.scoreddocs))
        assert len(list(dataset.scoreddocs)) == 10
        assert "8172" == list(dataset.scoreddocs)[0].doc_id
        assert "1" == list(dataset.scoreddocs)[0].query_id

    def test_for_scoreddocs_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        dataset = ir_datasets.load("tirex-sample-dataset")
        print(list(dataset.scoreddocs))
        assert len(list(dataset.scoreddocs)) == 10
        assert "8172" == list(dataset.scoreddocs)[0].doc_id
        assert "1" == list(dataset.scoreddocs)[0].query_id

    def test_for_scoreddocs_within_tira(self):
        os.environ["TIRA_INPUT_DATASET"] = "tests/resources/re-ranking-outputs"
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        dataset = ir_datasets.load("msmarco-passage/eval/small")
        print(list(dataset.scoreddocs))
        del os.environ["TIRA_INPUT_DATASET"]
        assert len(list(dataset.scoreddocs)) == 3
        assert "doc-1" == list(dataset.scoreddocs)[0].doc_id
        assert "1" == list(dataset.scoreddocs)[0].query_id
        assert 10 == list(dataset.scoreddocs)[0].score
        register_rerank_data_to_ir_datasets("tests/sample-tirex-rerank-data", "tirex-sample-dataset", None)

    def test_loading_raw_ir_datasets_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        ir_datasets = load_ir_datasets()
        dataset = ir_datasets.load("cranfield")
        queries = {str(i.query_id): i.text for i in dataset.queries_iter()}

        assert len(list(dataset.queries_iter())) == 225
        assert queries["269"] == "has a criterion been established for determining the axial compressor\nchoking line ."

    def test_loading_raw_ir_datasets_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        ir_datasets = load_ir_datasets()
        dataset = ir_datasets.load("cranfield")
        queries = {str(i.query_id): i.text for i in dataset.queries_iter()}

        assert len(list(dataset.queries_iter())) == 225
        assert queries["269"] == "has a criterion been established for determining the axial compressor\nchoking line ."

    def test_loading_queries_from_ir_datasets_from_custom_directory_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        os.environ["TIRA_INPUT_DATASET"] = "tests/resources/sample-input-full-rank"
        ir_datasets = load_ir_datasets()
        del os.environ["TIRA_INPUT_DATASET"]
        dataset = ir_datasets.load("does-not-exist-and-is-not-used")
        queries = {i.query_id: i.text for i in dataset.queries_iter()}

        assert len(list(dataset.queries_iter())) == 2
        assert queries["1"] == "fox jumps above animal"
        self.assertEqual("static_ir_dataset-tests/resources/sample-input-full-rank", dataset.dataset_id())

    def test_loading_queries_from_ir_datasets_from_custom_directory_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        os.environ["TIRA_INPUT_DATASET"] = "tests/resources/sample-input-full-rank"
        ir_datasets = load_ir_datasets()
        del os.environ["TIRA_INPUT_DATASET"]
        dataset = ir_datasets.load("does-not-exist-and-is-not-used")
        queries = {i.query_id: i.text for i in dataset.queries_iter()}

        assert len(list(dataset.queries_iter())) == 2
        assert queries["1"] == "fox jumps above animal"

    def test_loading_queries_from_ir_datasets_from_custom_directory_2_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        os.environ["TIRA_INPUT_DATASET"] = "tests/resources/sample-input-full-rank-without-documents"
        ir_datasets = load_ir_datasets()
        del os.environ["TIRA_INPUT_DATASET"]
        dataset = ir_datasets.load("does-not-exist-and-is-not-used")
        queries = {i.query_id: i.text for i in dataset.queries_iter()}

        assert len(list(dataset.queries_iter())) == 2
        assert queries["1"] == "fox jumps above animal"

    def test_loading_queries_from_ir_datasets_from_custom_directory_2_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        os.environ["TIRA_INPUT_DATASET"] = "tests/resources/sample-input-full-rank-without-documents"
        ir_datasets = load_ir_datasets()
        del os.environ["TIRA_INPUT_DATASET"]
        dataset = ir_datasets.load("does-not-exist-and-is-not-used")
        queries = {i.query_id: i.text for i in dataset.queries_iter()}

        assert len(list(dataset.queries_iter())) == 2
        assert queries["1"] == "fox jumps above animal"

    def test_loading_documents_from_ir_datasets_from_custom_directory_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        os.environ["TIRA_INPUT_DATASET"] = "tests/resources/sample-input-full-rank"
        ir_datasets = load_ir_datasets()
        del os.environ["TIRA_INPUT_DATASET"]
        dataset = ir_datasets.load("does-not-exist-and-is-not-used")
        docs = {i.doc_id: i.text for i in dataset.docs_iter()}

        assert len(list(dataset.docs_iter())) == 5
        assert docs["pangram-02"] == "Quick fox jumps nightly above wizard."

    def test_loading_documents_from_ir_datasets_from_custom_directory_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        os.environ["TIRA_INPUT_DATASET"] = "tests/resources/sample-input-full-rank"
        ir_datasets = load_ir_datasets()
        del os.environ["TIRA_INPUT_DATASET"]
        dataset = ir_datasets.load("does-not-exist-and-is-not-used")
        docs = {i.doc_id: i.text for i in dataset.docs_iter()}

        assert len(list(dataset.docs_iter())) == 5
        assert docs["pangram-02"] == "Quick fox jumps nightly above wizard."

    def test_loading_docs_store_from_ir_datasets_from_custom_directory_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        os.environ["TIRA_INPUT_DATASET"] = "tests/resources/sample-input-full-rank"
        ir_datasets = load_ir_datasets()
        del os.environ["TIRA_INPUT_DATASET"]
        dataset = ir_datasets.load("does-not-exist-and-is-not-used")

        assert dataset.docs_store()["pangram-03"].text == "The jay, pig, fox, zebra and my wolves quack!"

    def test_loading_docs_store_from_ir_datasets_from_custom_directory_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        os.environ["TIRA_INPUT_DATASET"] = "tests/resources/sample-input-full-rank"
        ir_datasets = load_ir_datasets()
        del os.environ["TIRA_INPUT_DATASET"]
        dataset = ir_datasets.load("does-not-exist-and-is-not-used")

        assert dataset.docs_store()["pangram-03"].text == "The jay, pig, fox, zebra and my wolves quack!"

    def test_no_patching_for_pyterrier_datasets_01(self):
        ensure_pyterrier_is_loaded()
        import pyterrier as pt

        dataset = pt.get_dataset("irds:cranfield")
        queries = {str(i["qid"]): i["query"] for _, i in dataset.get_topics().iterrows()}

        assert len(dataset.get_topics()) == 225
        assert queries["269"] == "has a criterion been established for determining the axial compressor choking line"

    def test_no_patching_for_pyterrier_datasets_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        import pyterrier as pt

        dataset = pt.get_dataset("irds:cranfield")
        queries = {str(i["qid"]): i["query"] for _, i in dataset.get_topics().iterrows()}

        assert len(dataset.get_topics()) == 225
        assert queries["269"] == "has a criterion been established for determining the axial compressor choking line"

    def test_patching_for_pyterrier_datasets_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        os.environ["TIRA_INPUT_DATASET"] = "tests/resources/sample-input-full-rank"
        import ir_datasets

        ir_load = ir_datasets.load
        ensure_pyterrier_is_loaded()
        import pyterrier as pt

        dataset = pt.get_dataset("irds:does-not-exist-and-is-not-used")
        ir_datasets.load = ir_load
        del os.environ["TIRA_INPUT_DATASET"]

        queries = {i["qid"]: i["query"] for _, i in dataset.get_topics().iterrows()}

        assert len(list(dataset.get_topics("title"))) == 2
        assert queries["1"] == "fox jumps above animal"

    def test_patching_for_pyterrier_datasets_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        os.environ["TIRA_INPUT_DATASET"] = "tests/resources/sample-input-full-rank"
        import ir_datasets

        ir_load = ir_datasets.load
        ensure_pyterrier_is_loaded()
        import pyterrier as pt

        dataset = pt.get_dataset("irds:does-not-exist-and-is-not-used")
        ir_datasets.load = ir_load
        del os.environ["TIRA_INPUT_DATASET"]

        queries = {i["qid"]: i["query"] for _, i in dataset.get_topics().iterrows()}

        assert len(list(dataset.get_topics("title"))) == 2
        assert queries["1"] == "fox jumps above animal"

    def test_patching_for_pyterrier_datasets_for_text_queries_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        os.environ["TIRA_INPUT_DATASET"] = "tests/resources/sample-input-full-rank"
        import ir_datasets

        ir_load = ir_datasets.load
        ensure_pyterrier_is_loaded()
        import pyterrier as pt

        dataset = pt.get_dataset("irds:does-not-exist-and-is-not-used")
        ir_datasets.load = ir_load
        del os.environ["TIRA_INPUT_DATASET"]

        queries = {i["qid"]: i["query"] for _, i in dataset.get_topics("text").iterrows()}

        assert len(list(dataset.get_topics("text"))) == 2
        assert queries["1"] == "fox jumps above animal"

    def test_patching_for_pyterrier_datasets_for_text_queries_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        os.environ["TIRA_INPUT_DATASET"] = "tests/resources/sample-input-full-rank"
        import ir_datasets

        ir_load = ir_datasets.load
        ensure_pyterrier_is_loaded()
        import pyterrier as pt

        dataset = pt.get_dataset("irds:does-not-exist-and-is-not-used")
        ir_datasets.load = ir_load
        del os.environ["TIRA_INPUT_DATASET"]

        queries = {i["qid"]: i["query"] for _, i in dataset.get_topics("text").iterrows()}

        assert len(list(dataset.get_topics("text"))) == 2
        assert queries["1"] == "fox jumps above animal"

    def test_patching_for_pyterrier_datasets_for_title_queries_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        os.environ["TIRA_INPUT_DATASET"] = "tests/resources/sample-input-full-rank"
        import ir_datasets

        ir_load = ir_datasets.load
        ensure_pyterrier_is_loaded()
        import pyterrier as pt

        dataset = pt.get_dataset("irds:does-not-exist-and-is-not-used")
        ir_datasets.load = ir_load
        del os.environ["TIRA_INPUT_DATASET"]

        queries = {i["qid"]: i["query"] for _, i in dataset.get_topics("title").iterrows()}

        assert len(list(dataset.get_topics("title"))) == 2
        assert queries["1"] == "fox jumps above animal"

    def test_patching_for_pyterrier_datasets_for_title_queries_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        os.environ["TIRA_INPUT_DATASET"] = "tests/resources/sample-input-full-rank"
        import ir_datasets

        ir_load = ir_datasets.load
        ensure_pyterrier_is_loaded()
        import pyterrier as pt

        dataset = pt.get_dataset("irds:does-not-exist-and-is-not-used")
        ir_datasets.load = ir_load
        del os.environ["TIRA_INPUT_DATASET"]

        queries = {i["qid"]: i["query"] for _, i in dataset.get_topics("title").iterrows()}

        assert len(list(dataset.get_topics("title"))) == 2
        assert queries["1"] == "fox jumps above animal"

    def test_loading_ir_dataset_via_rest_api_from_tira_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        from tira.third_party_integrations import ir_datasets

        dataset = ir_datasets.load("ir-lab-jena-leipzig-sose-2023/iranthology-20230618-training")

        assert dataset.has_docs(), "dataset has no documents"
        assert dataset.has_queries(), "dataset has no queries"
        assert dataset.has_qrels(), "dataset has no qrels"

    def test_loading_ir_dataset_via_rest_api_from_tira_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        from tira.third_party_integrations import ir_datasets

        dataset = ir_datasets.load("ir-lab-jena-leipzig-sose-2023/iranthology-20230618-training")

        assert dataset.has_docs(), "dataset has no documents"
        assert dataset.has_queries(), "dataset has no queries"
        assert dataset.has_qrels(), "dataset has no qrels"

    def test_loading_qrels_via_rest_api_from_tira_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        from tira.third_party_integrations import ir_datasets

        dataset = ir_datasets.load("ir-lab-jena-leipzig-sose-2023/iranthology-20230618-training")

        assert len([i for i in dataset.qrels_iter()]) == 2635
        assert (
            "TrecQrel(query_id='1', doc_id='2005.ipm_journal-ir0anthology0volumeA41A1.7', relevance=1, iteration='0')"
            == str([i for i in dataset.qrels_iter()][0])
        )

    def test_loading_qrels_via_rest_api_from_tira_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        from tira.third_party_integrations import ir_datasets

        dataset = ir_datasets.load("ir-lab-jena-leipzig-sose-2023/iranthology-20230618-training")

        assert len([i for i in dataset.qrels_iter()]) == 2635
        assert (
            "TrecQrel(query_id='1', doc_id='2005.ipm_journal-ir0anthology0volumeA41A1.7', relevance=1, iteration='0')"
            == str([i for i in dataset.qrels_iter()][0])
        )

    def test_loading_docs_via_rest_api_from_tira_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        from tira.third_party_integrations import ir_datasets

        dataset = ir_datasets.load("ir-lab-jena-leipzig-sose-2023/iranthology-20230618-training")

        assert len([i for i in dataset.docs_iter()]) == 53673
        docsstore = dataset.docs_store()

        assert docsstore.get("2005.ipm_journal-ir0anthology0volumeA41A1.7").text.startswith(
            "A probabilistic model for stemmer generation AbstractIn this paper"
        )

    def test_loading_docs_via_rest_api_from_tira_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        from tira.third_party_integrations import ir_datasets

        dataset = ir_datasets.load("ir-lab-jena-leipzig-sose-2023/iranthology-20230618-training")

        assert len([i for i in dataset.docs_iter()]) == 53673
        docsstore = dataset.docs_store()

        assert docsstore.get("2005.ipm_journal-ir0anthology0volumeA41A1.7").text.startswith(
            "A probabilistic model for stemmer generation AbstractIn this paper"
        )

    def test_loading_topics_via_rest_api_from_tira_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        from tira.third_party_integrations import ir_datasets

        dataset = ir_datasets.topics_file("ir-lab-jena-leipzig-wise-2023/training-20231104-training")

        self.assertIn("/training-20231104-training/truth-data/queries.xml", str(dataset))

    def test_loading_topics_via_rest_api_from_tira_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        from tira.third_party_integrations import ir_datasets

        topics_file = ir_datasets.topics_file("ir-lab-jena-leipzig-wise-2023/training-20231104-training")

        self.assertIn("/training-20231104-training/truth-data/queries.xml", str(topics_file))

    def test_loading_topics_via_rest_api_from_rerank_dataset_from_tira_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        os.environ["TIRA_INPUT_DATASET"] = "tests/resources/re-ranking-outputs"
        ir_datasets = load_ir_datasets()
        dataset = ir_datasets.load("workshop-on-open-web-search/re-ranking-20231027-training")
        del os.environ["TIRA_INPUT_DATASET"]

        assert len([i for i in dataset.queries_iter()]) == 2
        assert len([i for i in dataset.docs_iter()]) == 3

    def test_loading_queries_via_rest_api_from_tira_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        from tira.third_party_integrations import ir_datasets

        dataset = ir_datasets.load("ir-lab-jena-leipzig-sose-2023/iranthology-20230618-training")

        assert 68 == len(list(dataset.queries_iter()))
        print(str(list(dataset.queries_iter())[0]))
        assert (
            "TirexQuery(query_id='1', text='retrieval system improving effectiveness', title='retrieval system"
            " improving effectiveness', query='retrieval system improving effectiveness', description='What papers"
            " focus on improving the effectiveness of a retrieval system?', narrative='Relevant papers include research"
            " on what makes a retrieval system effective and what improves the effectiveness of a retrieval system."
            " Papers that focus on improving something else or improving the effectiveness of a system that is not a"
            " retrieval system are not relevant.')" == str(list(dataset.queries_iter())[0])
        )

    def test_loading_queries_via_rest_api_from_tira_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        from tira.third_party_integrations import ir_datasets

        dataset = ir_datasets.load("ir-lab-jena-leipzig-sose-2023/iranthology-20230618-training")

        assert 68 == len(list(dataset.queries_iter()))
        print(str(list(dataset.queries_iter())[0]))
        assert (
            "TirexQuery(query_id='1', text='retrieval system improving effectiveness', title='retrieval system"
            " improving effectiveness', query='retrieval system improving effectiveness', description='What papers"
            " focus on improving the effectiveness of a retrieval system?', narrative='Relevant papers include research"
            " on what makes a retrieval system effective and what improves the effectiveness of a retrieval system."
            " Papers that focus on improving something else or improving the effectiveness of a system that is not a"
            " retrieval system are not relevant.')" == str(list(dataset.queries_iter())[0])
        )

    def test_patching_for_pyterrier_datasets_to_tira(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        import pyterrier as pt

        dataset = pt.get_dataset("irds:workshop-on-open-web-search/retrieval-20231027-training")
        queries = {i["qid"]: i["title"] for _, i in dataset.get_topics().iterrows()}

        assert len(dataset.get_topics()) == 3
        assert queries["1"] == "hubble telescope achievements"

    def test_patching_for_pyterrier_datasets_with_qrels_to_tira(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        import pyterrier as pt

        dataset = pt.get_dataset("irds:workshop-on-open-web-search/retrieval-20231027-training")
        qrels = [i.to_dict() for _, i in dataset.get_qrels().iterrows()]

        assert len(dataset.get_qrels()) == 3
        assert qrels[0] == {"docno": "doc-1", "iteration": "0", "label": 1, "qid": "1"}

    def test_ir_datasets_id_is_available(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        import pyterrier as pt

        expected = "workshop-on-open-web-search/retrieval-20231027-training"

        dataset = pt.get_dataset("irds:workshop-on-open-web-search/retrieval-20231027-training")
        actual = dataset.irds_ref().dataset_id()

        self.assertEqual(expected, actual)

    def test_ir_datasets_get_many_iter_is_available_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        import pyterrier as pt

        expected = []

        dataset = pt.get_dataset("irds:workshop-on-open-web-search/retrieval-20231027-training")
        actual = [i for i in dataset.irds_ref().docs_store().get_many_iter(["1", "2"]) if i]

        self.assertEqual(expected, actual)

    def test_ir_datasets_get_many_iter_is_available_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        import pyterrier as pt

        expected = ["doc-2", "doc-3"]

        dataset = pt.get_dataset("irds:workshop-on-open-web-search/retrieval-20231027-training")
        actual = [i.doc_id for i in dataset.irds_ref().docs_store().get_many_iter(["doc-2", "doc-3"]) if i]

        self.assertEqual(expected, actual)



import unittest

import pandas as pd

from tira.local_client import Client

# .. todo:: this file still uses "unclean" assertions and should be converted to use unittest assertions


def retrieval_submission(queries):
    queries = pd.DataFrame([{"qid": str(i)} for i in queries])
    tira = Client("tests/resources/")
    retriever = tira.pt.from_retriever_submission(
        "ir-benchmarks/tira-ir-starters/retriever-for-re-ranking", dataset="d1"
    )

    ret = retriever(queries)
    return [{"qid": i["qid"], "docno": i["docno"], "score": i["score"], "text": i["text"]} for _, i in ret.iterrows()]


class TestPtFromRetrieverForReRankingSubmissionLoader(unittest.TestCase):
    def test_for_multiple_queries(self):
        expected = [
            {"qid": "1", "docno": "doc-1", "score": 10, "text": "Text of doc-1"},
            {"qid": "1", "docno": "doc-2", "score": 9, "text": "Text of doc-2"},
            {"qid": "3", "docno": "doc-3", "score": 1, "text": "Text of doc-3"},
        ]

        actual = retrieval_submission([1, 3])

        assert len(actual) == 3
        assert actual[0] == expected[0]
        assert actual[1] == expected[1]
        assert actual[2] == expected[2]

    def test_for_single_query_no_1(self):
        expected = [
            {"qid": "1", "docno": "doc-1", "score": 10, "text": "Text of doc-1"},
            {"qid": "1", "docno": "doc-2", "score": 9, "text": "Text of doc-2"},
        ]

        actual = retrieval_submission([1])

        assert len(actual) == 2
        assert actual[0] == expected[0]
        assert actual[1] == expected[1]

    def test_for_single_query_no_3(self):
        expected = [
            {"qid": "3", "docno": "doc-3", "score": 1, "text": "Text of doc-3"},
        ]

        actual = retrieval_submission([3])

        assert len(actual) == 1
        assert actual[0] == expected[0]



import unittest

from tira.io_utils import change_workdir_cmd


class ChangeWorkdirTest(unittest.TestCase):
    def test_none_returns_empty_chdir_cmd(self):
        expected = 'echo "did not change the working directory"'
        actual = change_workdir_cmd(None)

        self.assertEqual(actual, expected)

    def test_empty_string_returns_empty_chdir_cmd(self):
        expected = 'echo "did not change the working directory"'
        actual = change_workdir_cmd("")

        self.assertEqual(actual, expected)

    def test_string_with_only_whitespaces_returns_empty_chdir_cmd(self):
        expected = 'echo "did not change the working directory"'
        actual = change_workdir_cmd("    ")

        self.assertEqual(actual, expected)

    def test_string_with_non_absolute_foo_returns_empty_chdir_cmd(self):
        expected = 'echo "did not change the working directory"'
        actual = change_workdir_cmd("   hello world ")

        self.assertEqual(actual, expected)

    def test_string_correct_string_01(self):
        expected = 'cd /app; echo "changed workdir to /app"'
        actual = change_workdir_cmd("/app")

        self.assertEqual(actual, expected)

    def test_string_correct_string_02(self):
        expected = 'cd /app/; echo "changed workdir to /app/"'
        actual = change_workdir_cmd("/app/")

        self.assertEqual(actual, expected)

    def test_string_correct_string_03(self):
        expected = 'cd /app/; echo "changed workdir to /app/"'
        actual = change_workdir_cmd("/app/   ")

        self.assertEqual(actual, expected)

    def test_string_correct_string_04(self):
        expected = 'cd /hello/world; echo "changed workdir to /hello/world"'
        actual = change_workdir_cmd("   /hello/world ")

        self.assertEqual(actual, expected)



import json
import tempfile
import unittest
from pathlib import Path

from approvaltests import verify_as_json

from tira.check_format import check_format, lines_if_valid

from . import _ERROR, _OK, EMPTY_OUTPUT, IR_QUERY_OUTPUT, JSONL_GZ_OUTPUT_VALID, TOUCHE_IMAGE_RETRIEVAL


class TestQueryProcessorFormat(unittest.TestCase):
    def test_error_message_on_empty_output(self):
        expected = [_ERROR, "No unique *.jsonl file was found, only the files ['.gitkeep'] were available."]
        actual = check_format(EMPTY_OUTPUT, "touche-image-retrieval")
        self.assertEqual(expected, actual)

    def test_error_message_on_ir_query_output(self):
        expected = "The json line misses the required field"
        actual = check_format(IR_QUERY_OUTPUT, "touche-image-retrieval")
        self.assertEqual(_ERROR, actual[0])
        self.assertIn(expected, actual[1])

    def test_error_message_on_json_gz_output(self):
        expected = "The json line misses the required field"
        actual = check_format(JSONL_GZ_OUTPUT_VALID, "touche-image-retrieval")
        self.assertEqual(_ERROR, actual[0])
        self.assertIn(expected, actual[1])

    def test_error_message_on_valid_output(self):
        expected = "The jsonl file has the correct format."
        actual = check_format(TOUCHE_IMAGE_RETRIEVAL, "touche-image-retrieval")
        self.assertEqual(_OK, actual[0])
        self.assertIn(expected, actual[1])

    def test_lines_for_valid_output(self):
        expected = {
            "qid": "1-5",
            "docno": "I002e616104f6ec04fd1a24d5",
            "q0": "0",
            "rank": 1,
            "score": 999,
            "system": "touche organizers - example submission for image retrieval; manual selection of images",
        }
        actual = lines_if_valid(TOUCHE_IMAGE_RETRIEVAL, "touche-image-retrieval")
        self.assertEqual(3, len(actual))
        self.assertEqual(expected, actual[0])



import unittest

from tira.local_execution_integration import LocalExecutionIntegration as Client


class TestLocalExecutionIntegrationTest(unittest.TestCase):
    def test_01(self):
        expected = "registry.webis.de/code-research/tira/tira-user-dam/seupd2324:latest"
        actual = Client().normalize_image_name("seupd2324", "registry.webis.de/code-research/tira/tira-user-dam/")

        self.assertEqual(expected, actual)

    def test_02(self):
        expected = "registry.webis.de/code-research/tira/tira-user-dam/seupd2324:latest"
        actual = Client().normalize_image_name(
            "seupd2324-dam-baseline", "registry.webis.de/code-research/tira/tira-user-dam/"
        )

        self.assertEqual(expected, actual)

    def test_03(self):
        expected = "registry.webis.de/code-research/tira/tira-user-petropoulossiblings/panospetr:1.1.2"
        actual = Client().normalize_image_name(
            "panospetr/av-software-petropoulos-pan2024:1.1.2",
            "registry.webis.de/code-research/tira/tira-user-petropoulossiblings/",
        )

        self.assertEqual(expected, actual)

    def test_04(self):
        expected = "registry.webis.de/code-research/tira/tira-user-petropoulossiblings/bash:latest"
        actual = Client().normalize_image_name(
            "bash:latest", "registry.webis.de/code-research/tira/tira-user-petropoulossiblings/"
        )

        self.assertEqual(expected, actual)
